# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""CUDA Toolkit loader/wrapper.
"""

import ctypes
import glob
import os
from sys import platform

from . import verbosity


def _find_cuda_path(path):
    n = len(path)
    i = 0
    while i < n:
        cpath = path[i:]
        j = cpath.find('CUDA')
        if j < 0:
            return None
        f = cpath[: j].rfind(';')
        if f < 0:
            f = 0
        else:
            f += 1
        t = cpath[j :].find(';')
        if t < 0:
            t = len(cpath)
        else:
            t += j
        b = cpath[f : t].find('bin')
        if b < 0:
            i += t
            continue
        return cpath[f : t]
    return None


try:
    if platform == 'win32':
        cuda_path = _find_cuda_path(os.environ['PATH'])
        cudart_dll = glob.glob(cuda_path + '/cudart64*')[0]
        cuda = ctypes.CDLL(cudart_dll, mode=ctypes.RTLD_GLOBAL)
    else:
        cuda = ctypes.CDLL('libcudart.so', mode=ctypes.RTLD_GLOBAL)

    v = ctypes.c_int()
    cuda.cudaRuntimeGetVersion(ctypes.byref(v))
    version = v.value
    if verbosity.level > 0:
        print('CUDA version: %d' % version)
except:
    if verbosity.level > 0:
        print('CUDA Toolkit not found, switching to cpu...')
    raise RuntimeError('CUDA Toolkit not found')


if version >= 10000:
    UUID_SIZE = 32
else:
    UUID_SIZE = 0

POINTER = ctypes.POINTER


class CudaDeviceProp(ctypes.Structure):
    _fields_ = [ 
            ('name', ctypes.c_char * 256),
            ('uuid', ctypes.c_char * UUID_SIZE),
            ('totalGlobalMem', ctypes.c_size_t),
            ('sharedMemPerBlock', ctypes.c_size_t),
            ('regsPerBlock', ctypes.c_int),
            ('warpSize', ctypes.c_int),
            ('memPitch', ctypes.c_size_t),
            ('maxThreadsPerBlock', ctypes.c_int),
            ('maxThreadsDim', ctypes.c_int * 3), 
            ('maxGridSize', ctypes.c_int * 3), 
            ('clockRate', ctypes.c_int),
            ('totalConstMem', ctypes.c_size_t),
            ('major', ctypes.c_int),
            ('minor', ctypes.c_int),
            ('textureAlignment', ctypes.c_size_t),
            ('texturePitchAlignment', ctypes.c_size_t),
            ('deviceOverlap', ctypes.c_int),
            ('multiProcessorCount', ctypes.c_int),
            ('kernelExecTimeoutEnabled', ctypes.c_int),
            ('integrated', ctypes.c_int),
            ('canMapHostMemory', ctypes.c_int),
            ('computeMode', ctypes.c_int),
            ('maxTexture1D', ctypes.c_int),
            ('maxTexture1DMipmap', ctypes.c_int),
            ('maxTexture1DLinear', ctypes.c_int),
            ('maxTexture2D', ctypes.c_int * 2), 
            ('maxTexture2DMipmap', ctypes.c_int * 2), 
            ('maxTexture2DLinear', ctypes.c_int * 3), 
            ('maxTexture2DGather', ctypes.c_int * 2), 
            ('maxTexture3D', ctypes.c_int * 3), 
            ('maxTexture3DAlt', ctypes.c_int * 3), 
            ('maxTextureCubemap', ctypes.c_int),
            ('maxTexture1DLayered', ctypes.c_int * 2), 
            ('maxTexture2DLayered', ctypes.c_int * 3), 
            ('maxTextureCubemapLayered', ctypes.c_int * 2), 
            ('maxSurface1D', ctypes.c_int),
            ('maxSurface2D', ctypes.c_int * 2), 
            ('maxSurface3D', ctypes.c_int * 3), 
            ('maxSurface1DLayered', ctypes.c_int * 2), 
            ('maxSurface2DLayered', ctypes.c_int * 3), 
            ('maxSurfaceCubemap', ctypes.c_int),
            ('maxSurfaceCubemapLayered', ctypes.c_int * 2), 
            ('surfaceAlignment', ctypes.c_size_t),
            ('concurrentKernels', ctypes.c_int),
            ('ECCEnabled', ctypes.c_int),
            ('pciBusID', ctypes.c_int),
            ('pciDeviceID', ctypes.c_int),
            ('pciDomainID', ctypes.c_int),
            ('tccDriver', ctypes.c_int),
            ('asyncEngineCount', ctypes.c_int),
            ('unifiedAddressing', ctypes.c_int),
            ('memoryClockRate', ctypes.c_int),
            ('memoryBusWidth', ctypes.c_int),
            ('l2CacheSize', ctypes.c_int),
            ('maxThreadsPerMultiProcessor', ctypes.c_int),
            ('streamPrioritiesSupported', ctypes.c_int),
            ('globalL1CacheSupported', ctypes.c_int),
            ('localL1CacheSupported', ctypes.c_int),
            ('sharedMemPerMultiprocessor', ctypes.c_size_t),
            ('regsPerMultiprocessor', ctypes.c_int),
            ('managedMemSupported', ctypes.c_int),
            ('isMultiGpuBoard', ctypes.c_int),
            ('multiGpuBoardGroupID', ctypes.c_int),
            ('singleToDoublePrecisionPerfRatio', ctypes.c_int),
            ('pageableMemoryAccess', ctypes.c_int),
            ('concurrentManagedAccess', ctypes.c_int),
            ('rest', ctypes.c_int * 1024)
            ]


getDeviceCount = cuda.cudaGetDeviceCount
getDeviceProperties = cuda.cudaGetDeviceProperties
synchronize = cuda.cudaDeviceSynchronize
malloc = cuda.cudaMalloc
malloc.argtypes = [POINTER(POINTER(ctypes.c_ubyte)), ctypes.c_int]
malloc.restype = ctypes.c_int
free = cuda.cudaFree
free.argtypes = [ctypes.c_void_p]
free.restype = ctypes.c_int
memset = cuda.cudaMemset
memset.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
memset.restype = ctypes.c_int
memcpy = cuda.cudaMemcpy
memcpy.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
memcpy.restype = ctypes.c_int
memcpy2D = cuda.cudaMemcpy2D
memcpy2D.restype = ctypes.c_int
memcpyH2H = ctypes.c_int(0)
memcpyH2D = ctypes.c_int(1)
memcpyD2H = ctypes.c_int(2)
memcpyD2D = ctypes.c_int(3)
createStream = cuda.cudaStreamCreate
createStream.restype = ctypes.c_int
deleteStream = cuda.cudaStreamDestroy

numDevices = ctypes.c_int()
getDeviceCount(ctypes.byref(numDevices))
if verbosity.level > 1:
    print('devices found: %d' % numDevices.value)
    for x in range(numDevices.value):
        devProp = CudaDeviceProp()
        getDeviceProperties(ctypes.byref(devProp), x)
        print('device: %s' % devProp.name)
        print('memory: %d' % devProp.totalGlobalMem)
        print('multiprocessors: %d' % devProp.multiProcessorCount)
        print('threads per mp: %d' % devProp.maxThreadsPerMultiProcessor)
        print('registers per block: %d'% devProp.regsPerBlock)
        print('shared memory per block: %d' % devProp.sharedMemPerBlock)



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

# -*- coding: utf-8 -*-
"""Principal Component Analysis of a dataset represented by a 2D ndarray.
"""

import numpy
import numpy.linalg as nla

from ..core.solver import Options
from ..algebra.dense_matrix import AMatrix
from .lra import LowerRankApproximation


def pca(A, npc=-1, tol=0, have=None, batch_size=None, verb=0, arch='cpu', \
        norm='f', mpc=-1, svtol=1e-3, opt=Options()):
    '''Performs principal component analysis for the set of data items
    represented by rows of a dense matrix A.

    For a given m-by-n data matrix A (m data samples n features each) computes
    m-by-k matrix L and k-by-n matrix R with k not greater than m and n such
    that the product L R approximates A_s = A - e a, where e = numpy.ones((m, 1)),
    and a = numpy.mean(A, axis=0).reshape((1, n))

    The rows of R (principal components) are orhonormal, the columns of L
    (reduced-features data) are orthogonal and are in the descending order of
    their norms.

    Parameters: basic
    -----------------
    A : 2D numpy array
        Data matrix.
    npc : int
        Required number of principal components if known.
        If negative, implicitely defined by the required accuracy of
        PCA approximation or interactively by the user.
    tol : float
        Approximation tolerance in the case npc < 0. If tol is non-zero, then 
        the computation of principal components will stop when the norm of 
        D = A_s - L R becomes not greater than eps, where eps is the norm
        of A_s multiplied by tol if tol > 0 and eps = -tol if tol < 0. Otherwise
        the user will be asked repeatedly whether the computation should 
        continue (the number of computed principal components and the relative 
        truncation error, the ratio of the norm of D to that of A_s, are
        displayed). Ignored if npc is positive.
    mpc : int
        Maximal number of PCs to compute. Ignored if non-positive, otherwise
        if mpc < min(m, n), then the accuracy of approximation set by tol
        might not be achieved. Ignored if npc is positive.
    have : tuple (a0, L0, R0)
        If not None, previously computed PCA approximation L0 R0 + e0 a0
        to a data matrix A0 of the same width as A is to be updated, i.e.
        PCA approximation L R + e a to numpy.concatenate((A0, A)) is to be
        computed. If neither nps nor tol are specified, the same number of
        components as in R0 is to be computed.
    batch_size : int
        If not None, performs incremental PCA with specified batch size.
        Principal components for the first batch are computed by the standard
        SVD, for each of the rest, update is performed (see 'have' above).
    verb : int
        Verbosity level.
    arch : string
        'cpu' : run on CPU,
        'gpu' : run on GPU if available, otherwise on CPU,
        'gpu!' : run on GPU, throw RuntimError if GPU is not present.

    Parameters: advanced
    --------------------
    norm : character
        The measure to be used for evaluating the relative approximation error:
        's' : the largest singular value,
        'f' : Frobenius norm,
        'm' : the largest norm of a row.
    svtol : float
        Error tolerance for singular values (see Notes below) relative to
        the largest singular value. May need to be decreased from its default
        value if highly accurate approximation is sought.
    opt : an object of class raleigh.solver.Options
        Solver options (see raleigh.solver).

    Returns
    -------
    mean : numpy array of shape (1, n)
        The mean of rows of A.
    trans : numpy array of shape (m, k)
        The reduced-features data matrix.
    comps : numpy array of shape (k, n)
        Principal components matrix.

    Usage examples
    --------------
    - Generate test data
    >>> import numpy
    >>> from raleigh.examples.pca.generate_matrix import generate
    >>> numpy.random.seed(1)
    >>> A, sigma, u, v = generate(3000, 2000, 1000, pca=True)

    - Compute 300 principal components of A:
    >>> mean, trans, comps = pca(A, npc=300)
    >>> em, ef = pca_error(A, mean, trans, comps)
    >>> print('PCA error: max 2-norm %.0e, Frobenius norm %.0e' % (em, ef))
    PCA error: max 2-norm 5e-02, Frobenius norm 1e-01

    - Compute a number of principal components delivering 5% relative
      accuracy of approximation of A:
    >>> mean, trans, comps = pca(A, tol=0.05)
    >>> em, ef = pca_error(A, mean, trans, comps)
    >>> print('PCA error: max 2-norm %.0e, Frobenius norm %.0e' % (em, ef))
    PCA error: max 2-norm 2e-02, Frobenius norm 4e-02

    - Compute PCA to 5% accuracy incrementally by processing 1000 data
      samples at a time:
    >>> mean, trans, comps = pca(A, batch_size=1000, tol=0.05)
    >>> em, ef = pca_error(A, mean, trans, comps)
    >>> print('PCA error: max 2-norm %.0e, Frobenius norm %.0e' % (em, ef))
    PCA error: max 2-norm 2e-02, Frobenius norm 4e-02

    - To demonstrate the use of update, let A[:2400, :] play the role of 'old'
      data A0
    >>> A0 = A[:2400, :]
    >>> mean, trans, comps = pca(A0, tol=0.05)
    >>> em, ef = pca_error(A0, mean, trans, comps)
    >>> print('PCA error: max 2-norm %.0e, Frobenius norm %.0e' % (em, ef))
    PCA error: max 2-norm 2e-02, Frobenius norm 5e-02

    - Now 'new' data A1 = A[2400:, :] arrived - update previously computed
      principal components:
    >>> A1 = A[2400:, :]
    >>> mean, trans, comps = pca(A1, have=(mean, trans, comps))
    >>> em, ef = pca_error(A, mean, trans, comps)
    >>> print('PCA error: max 2-norm %.0e, Frobenius norm %.0e' % (em, ef))
    PCA error: max 2-norm 2e-02, Frobenius norm 5e-02

    Notes
    -----
    If have is None, then the rows of R are approximate right singular
    vectors of A - e a and the columns of L are approximate left singular
    vectors of A - e a multiplied by respective singular values. If have is
    not None, this is generally not the case.
    '''
    lra = LowerRankApproximation(have)
    if batch_size is None:
        if have is None:
            data_matrix = AMatrix(A, arch=arch)
            lra.compute(data_matrix, opt=opt, rank=npc, tol=tol, norm=norm, \
                        max_rank=mpc, svtol=svtol, shift=True, verb=verb)
        else:
            data_matrix = AMatrix(A, arch=arch, copy_data=True)
            lra.update(data_matrix, opt=opt, rank=npc, tol=tol, norm=norm, \
                       max_rank=mpc, svtol=svtol, verb=verb)
    else:
        lra.icompute(A, batch_size, opt=opt, rank=npc, tol=tol, norm=norm, \
                     max_rank=mpc, svtol=svtol, shift=True, verb=verb, \
                     arch=arch)
    trans = lra.left()
    comps = lra.right()
    mean = lra.mean()
    return mean, trans, comps


def pca_error(data, mean, trans, comps):
    m, n = data.shape
    ones = numpy.ones((data.shape[0], 1), dtype=data.dtype)
    mean = numpy.reshape(mean, (1, comps.shape[1]))
    data_s = data - numpy.dot(ones, mean)
    err = numpy.dot(trans, comps) - data_s
    em = numpy.amax(_norm(err, axis=1))/numpy.amax(_norm(data_s, axis=1))
    ef = numpy.amax(nla.norm(err, ord='fro')) \
        /numpy.amax(nla.norm(data_s, ord='fro'))
    return em, ef


def _norm(a, axis):
    return numpy.apply_along_axis(nla.norm, axis, a)



'''Driver routines for RALEIGH solver. Provide user friendly SciPy-style
   interfaces to RALEIGH abstract solver.

lra:
    Implements a class for handling the computation of lower rank approximations
    of dense matrices.

partial_hevp:
    Computes several eigenpairs of sparse real symmetric/Hermitian eigenvalue
    problems using either shift-invert or preconditioning technique.
    Requires MKL 10.3 or later (needs mkl_rt.dll on Windows, libmkl_rt.so on
    Linux).

partial_svd:
    Computes several extreme singular values of a dense matrix and corresponding
    singular vectors. On CPU, performs better if MKL 10.3 or later is available.
    Can also run on GPU.

pca:
    Performs principal component analysis for the set of data items represented
    by rows of a dense matrix.

truncated_svd:
    Computes truncated Singular Value Decomposition of a dense matrix.
'''



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""Base classes for ndarray implementation of Vectors and Matrix types.
"""

import numbers
import numpy


class NDArrayVectors(object):
    '''Base class for ndarray-based Vectors type.
    '''

    '''========== Methods required by RALEIGH core solver ==========
    '''

    def dimension(self):
        return self.__vdim

    def nvec(self):
        return self.__selected[1]

    def select(self, nv, first=0):
        assert nv <= self.__nvec and first >= 0
        self.__selected = (first, nv)

    def selected(self):
        return self.__selected

    def data_type(self):
        return self.__dtype

    def fill_random(self):
        iv, nv = self.__selected
        m, n = self.__data.shape
        self.__data[iv : iv + nv, :] = 2*numpy.random.rand(nv, n) - 1

    def append(self, other, axis=0):
        if axis == 0:
            self.__data = numpy.concatenate((self.data(), other.data()))
            self.__nvec = self.nvec() + other.nvec()
            self.select_all()
        else:
            self.__data = numpy.concatenate((self.__data, other.all_data()), \
                                             axis=1)
            self.__vdim += other.dimension()

    '''========== Other methods ====================================
    '''

    def __init__(self, arg, nvec=0, data_type=None, shallow=False):
        if isinstance(arg, NDArrayVectors):
            i, n = arg.selected()
            if shallow:
                self.__data = arg.__data[i : i + n, :]
            else:
                self.__data = arg.__data[i : i + n, :].copy()
        elif isinstance(arg, NDArrayMatrix):
            if shallow:
                self.__data = arg.data()
            else:
                self.__data = arg.data().copy()
            if not self.__data.flags['C_CONTIGUOUS']:
                raise ValueError('Vectors data must be C_CONTIGUOUS')
        elif isinstance(arg, numpy.ndarray):
            self.__data = arg
        elif isinstance(arg, numbers.Number):
            if data_type is None: # use default data type
                self.__data = numpy.zeros((nvec, arg))
            else:
                self.__data = numpy.zeros((nvec, arg), dtype=data_type)
        else:
            raise ValueError \
                ('wrong argument %s in constructor' % repr(type(arg)))
        dt = self.__data.dtype.type
        dk = self.__data.dtype.kind
        self.__dtype = dt
        self.__is_complex = (dk == 'c')
        m, n = self.__data.shape
        self.__selected = (0, m)
        self.__vdim = n
        self.__nvec = m

    def select_all(self):
        self.select(self.__nvec)

    def is_complex(self):
        return self.__is_complex

    def zero(self):
        f, n = self.__selected;
        self.__data[f : f + n, :] = 0.0

    def fill(self, array_or_value):
        f, n = self.__selected;
        self.__data[f : f + n, :] = array_or_value

    def fill_orthogonal(self):
        iv, nv = self.__selected
        k, n = self.__data.shape
        if n < nv:
            raise ValueError('fill_orthogonal: too many vectors in the array')
        _fill_ndarray_with_orthogonal_vectors(self.__data[iv : iv + nv, :])

    def all_data(self):
        return self.__data

    def data(self, i=None):
        f, n = self.__selected
        if i is None:
            return self.__data[f : f + n, :]
        else:
            return self.__data[f + i, :]


class NDArrayMatrix:

    def __init__(self, arg):
        if isinstance(arg, NDArrayVectors):
            data = arg.data()
        elif isinstance(arg, numpy.ndarray):
            data = arg
        else:
            raise ValueError \
                ('wrong argument %s in Matrix constructor' % repr(type(arg)))
        self.__data = data
        self.__shape = data.shape
        self.__dtype = data.dtype.type

    def data(self):
        return self.__data

    def shape(self):
        return self.__shape

    def data_type(self):
        return self.__dtype

    def is_complex(self):
        return (self.__data.dtype.kind == 'c')


def _fill_ndarray_with_orthogonal_vectors(a):
    a.fill(0.0)
    m, n = a.shape
    a[0,0] = 1.0
    i = 1
    while 2*i < m:
        a[i : 2*i, :i] = a[:i, :i]
        a[:i, i : 2*i] = a[:i, :i]
        a[i : 2*i, i : 2*i] = -a[:i, :i]
        i *= 2
    k = i
    j = 2*i
    if j > n:
        for i in range(k, m):
            a[i, i] = 1.0
        return
    while j <= n:
        a[:k, i : j] = a[:k, :i]
        i, j = j, 2*j
    j = i//2
    a[k : m,   : j] = a[: (m - k), : j]
    a[k : m, j : i] = -a[: (m - k), j : i]



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""Truncated SVD (and, optionally, PCA) demo.

Usage:
  truncated_svd [--help | -h | options] <data>

Arguments:
  data  numpy .npy file containing the matrix.

Options:
  -a <arch>, --arch=<arch>   architecture [default: cpu]
  -d <svd> , --svd=<svd>     .npz file contaning singular values and vectors
                             of the data matrix (if available)
  -n <rank>, --rank=<rank>   truncated svd rank (negative: unknown a priori, 
                             thsh > 0 will be used as a criterion,
                             if thsh == 0, will run in interactive mode)
                             [default: -1]
  -s <thsh>, --thres=<thsh>  singular values threshold [default: 0.01]
  -t <vtol>, --vtol=<vtol>   singular vectors error tolerance [default: 1e-3]
  -v <verb>, --verb=<verb>   verbosity level [default: 0]
  -P, --pca                  compute also PCA
"""

try:
    from docopt import docopt
    __version__ = '0.1.0'
    have_docopt = True
except:
    have_docopt = False

import numpy
import numpy.linalg as nla
import sys
import time

from raleigh.algebra import verbosity
verbosity.level = 2

from raleigh.core.solver import Options
from raleigh.interfaces.truncated_svd import truncated_svd
from raleigh.interfaces.pca import pca


def norm(a, axis):
    return numpy.apply_along_axis(nla.norm, axis, a)


def vec_err(u, v):
    w = v.copy()
    q = numpy.dot(u.T, v)
    w = numpy.dot(u, q) - v
    s = norm(w, axis = 0)
    return s


if have_docopt:
    args = docopt(__doc__, version=__version__)
    filename = args['<data>']
    A = numpy.load(filename)
    arch = args['--arch']
    rank = int(args['--rank'])
    th = float(args['--thres'])
    tol = float(args['--vtol'])
    verb = int(args['--verb'])
    do_pca = args['--pca']
    svd_data = None
    filename = args['--svd']
    if filename is not None:
        svd_data = numpy.load(filename)
    else:
        svd_data = None
else:
    narg = len(sys.argv)
    if narg < 2 or sys.argv[1] == '-h' or sys.argv[1] == '--help':
        print('\nUsage:\n')
        print('python truncated_svd.py <data>')
        exit()
    filename = sys.argv[1]
    A = numpy.load(filename)
    arch = 'cpu' if narg < 3 else 'gpu!'
    print('\n=== docopt not found, using default options...\n')
    rank = -1
    th = 0.01
    tol = 1e-3
    verb = 0
    do_pca = True
    svd_data = None

numpy.random.seed(1) # make results reproducible

m = A.shape[0]
if len(A.shape) > 2:
    n = numpy.prod(A.shape[1:])
    A = numpy.reshape(A, (m, n))
else:
    n = A.shape[1]
dtype = A.dtype.type

print('\n--- solving with truncated_svd...\n')
start = time.time()
u, sigma, vt = truncated_svd(A, nsv=rank, tol=th, vtol=tol, arch=arch)
stop = time.time()
time_tsvd = stop - start
print('\ntruncated svd time: %.1e' % time_tsvd)

nsv = sigma.shape[0]
print('\n%d singular vectors computed' % nsv)
nsv = numpy.sum(sigma > th)
print('%d singular values above threshold' % nsv)
if nsv > 0 and svd_data is not None:
    sigma0 = svd_data['sigma']
    v0 = svd_data['right']
    nsv = min(nsv, sigma0.shape[0])
    err_vec = vec_err(v0[:,:nsv], vt.transpose()[:,:nsv])
    err_val = abs(sigma[:nsv] - sigma0[:nsv])/sigma0[0]
    print('\nmax singular vector error: %.1e' % numpy.amax(err_vec))
    print('max singular value error: %.1e' % numpy.amax(err_val))
D = A - numpy.dot(sigma[:nsv]*u[:, :nsv], vt[:nsv, :])
err = numpy.amax(norm(D, axis=1))/numpy.amax(norm(A, axis=1))
print('\ntruncation error %.1e' % err)

if do_pca:
    print('\n--- solving with pca...\n')
    '''
    Principal components do not have to be accurate singular vectors, hence pca
    can compute them faster than truncated_svd in default singular vector accuracy
    (it is this trick that makes scikit-learn pca much faster than much more
    accurate scipy svds).
    Note that PCA computes SVD for a shifted-to-zero-average matrix, which has 
    smaller singular values than sigma computed above by truncated SVD: e.g. 
    in the case of a matrix generated by generate_matrix.py, shifted matrix has
    singular values sigma[1], sigma[2], ..., hence below the threshold is
    increased accordingly.
    '''
    th *= sigma[0]/sigma[1]
    start = time.time()
    mean, trans, comps = pca(A, npc=rank, tol=th, norm='s', arch=arch, \
                             verb=verb)
    stop = time.time()
    time_pca = stop - start
    print('\npca time: %.1e' % time_pca)
    s = norm(trans, axis=0)
    e = numpy.ones((trans.shape[0], 1), dtype=dtype)
    D = A - numpy.dot(trans, comps) - numpy.dot(e, mean)
    err = numpy.amax(norm(D, axis=1))/numpy.amax(norm(A, axis=1))
    print('\npca error %.1e' % err)

print('\ndone')



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

# -*- coding: utf-8 -*-
"""
Incremental Principal Component Analysis demo.

Principal components are computed incrementally, processing a number 
(batch size) of data samples at a time.

For comparison, the same amount of principal components is computed also by 
scikit-learn IncrementalPCA.

Usage: incremental_pca <data_file> <batch_size> <tolerance> <max_pcs> [gpu]

Arguments:

data_file  : the name of the file containing data matrix
batch_size : batch size
tolerance  : approximation tolerance wanted
max_pcs    : maximal number of principal components to compute (<1: no limit)
gpu        : run raleigh pca on GPU if this argument is present
"""

import numpy
import sys
import timeit

try:
    from sklearn.decomposition import IncrementalPCA
    have_sklearn = True
except:
    have_sklearn = False

from raleigh.interfaces.pca import pca, pca_error


narg = len(sys.argv)
if narg < 5:
    usage = \
   'Usage: incremental_pca <data_file> <batch_size> <tolerance> <max_pcs> [gpu]'
    raise SystemExit(usage)
data = numpy.load(sys.argv[1], mmap_mode='r')
atol = float(sys.argv[3])
batch_size = int(sys.argv[2])
mpc = int(sys.argv[4])
arch = 'cpu' if narg < 6 else 'gpu!'

numpy.random.seed(1) # make results reproducible

dtype = data.dtype
m = data.shape[0]
n = data.shape[1]
if len(data.shape) > 2:
    n = numpy.prod(data.shape[1:])
    data = numpy.memmap.reshape(data, (m, n))

print('\n--- solving with raleigh pca...\n')
start = timeit.default_timer()
mean, trans, comps = pca(data, mpc=mpc, tol=atol, arch=arch, verb=1, \
    batch_size=batch_size)
elapsed = timeit.default_timer() - start
ncomp = comps.shape[0]
print('%d principal components computed in %.2e sec' % (ncomp, elapsed))
em, ef = pca_error(data, mean, trans, comps)
print('PCA error: max 2-norm %.1e, Frobenius norm %.1e' % (em, ef))

if have_sklearn and arch == 'cpu':
    npc = ncomp
    if npc > batch_size:
        batch_size = npc
    print('\n--- solving with sklearn.decomposition.IncrementalPCA...\n')
    start = timeit.default_timer()
    skl_ipca = IncrementalPCA(n_components=npc, batch_size=batch_size)
    skl_trans = skl_ipca.fit_transform(data)
    skl_comps = skl_ipca.components_
    skl_mean = skl_ipca.mean_
    stop = timeit.default_timer()
    elapsed = stop - start
    ncomp = skl_comps.shape[0]
    print('%d principal components computed in %.2e sec' % (ncomp, elapsed))
    em, ef = pca_error(data, skl_mean, skl_trans, skl_comps)
    print('PCA error: max 2-norm %.1e, Frobenius norm %.1e' % (em, ef))

print('\ndone')



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""Wrapper for MKL sparse symmetric/Hermitian matrices and solvers working with
   SciPy sparse matrices.
"""

import numpy
import scipy.sparse as scs

from .mkl_wrap import SparseSymmetricMatrix as SSM
from .mkl_wrap import ParDiSo as SSS
from .mkl_wrap import ILUT


class SparseSymmetricMatrix:

    def __init__(self, matrix):
        try:
            csr = matrix.csr()
        except:
            csr = scs.triu(matrix, format='csr')
            csr.sort_indices()
        a = csr.data
        ia = csr.indptr + 1
        ja = csr.indices + 1
        self.__csr = csr
        self.__ssm = SSM(a, ia, ja)
        self.__a = a
        self.__ia = ia
        self.__ja = ja

    def size(self):
        return self.__csr.shape[0]

    def data_type(self):
        return self.__a.dtype

    def csr(self):
        return self.__csr

    def apply(self, x, y):
        try:
            x = x.data()
            y = y.data()
        except:
            pass
        self.__ssm.dot(x, y)


class SparseSymmetricSolver:

    def __init__(self, dtype=numpy.float64, pos_def=False):
        self.__solver = SSS(dtype=dtype, pos_def=pos_def)
        self.__dtype = dtype

    def analyse(self, a, sigma=0, b=None):
        data = a.data
        if sigma != 0:
            if b is None:
                b = scs.eye(a.shape[0], dtype=a.data.dtype, format='csr')
            a_s = a - sigma * b
        else:
            a_s = a
        a_s = scs.triu(a_s, format='csr')
        a_s.sort_indices()
        ia = a_s.indptr + 1
        ja = a_s.indices + 1
        data = a_s.data
        self.__solver.analyse(data, ia, ja)
        self.__n = ia.shape[0] - 1
        self.__sigma = sigma

    def factorize(self):
        self.__solver.factorize()

    def solve(self, b, x):
        try:
            b = b.data()
            x = x.data()
        except:
            pass
        self.__solver.solve(b, x)

    def apply(self, b, x):
        self.solve(b, x)

    def inertia(self):
        return self.__solver.inertia()

    def size(self):
        return self.__n

    def data_type(self):
        return self.__dtype

    def sigma(self):
        return self.__sigma

    def solver(self):
        return self.__solver


class IncompleteLU:

    def __init__(self, matrix):
        matrix = matrix.tocsr().sorted_indices()
        a = matrix.data
        ia = matrix.indptr + 1
        ja = matrix.indices + 1
        self.__ilut = ILUT(a, ia, ja)

    def factorize(self, tol=1e-6, max_fill=1):
        self.__ilut.factorize(tol=tol, max_fill_rel=max_fill)

    def apply(self, x, y):
        try:
            x = x.data()
            y = y.data()
        except:
            pass
        self.__ilut.solve(x, y)


class Operator:

    def __init__(self, op):
        self.__op = op

    def apply(self, x, y):
        try:
            x = x.data()
            y = y.data()
        except:
            pass
        self.__op.apply(x, y)



'''RALEIGH core solver.

    Contains core eigensolver routine implemented in terms of an abstract
    Vectors class.

    Direct use is for advanced users only - start by trying user-friendly
    interfaces in raleigh/interfaces.
'''



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""Architecture-aware wrap for a dense matrix.
"""

import numpy


class AMatrix:

    def __init__(self, a, arch='cpu', copy_data=False):
        self.__arch = arch
        if arch[:3] == 'gpu':
            try:
                from . import cuda_wrap as cuda
                from .dense_cublas import Matrix, Vectors
                self.__op = Matrix(a)
                self.__gpu = cuda
            except:
                if len(arch) > 3 and arch[3] == '!':
                    raise RuntimeError('cannot use GPU')
        else:
            from .dense_cpu import Matrix, Vectors
#            from .dense_numpy import Matrix, Vectors
            if copy_data:
                self.__op = Matrix(a.copy())
            else:
                self.__op = Matrix(a)
            self.__gpu = None
        self.__Vectors = Vectors
        self.__vectors = None
        vmin = numpy.amin(a)
        vmax = numpy.amax(a)
        self.__scale = max(abs(vmin), abs(vmax))

    def as_operator(self):
        return self.__op

    def as_vectors(self):
        if self.__vectors is None:
            self.__vectors = self.__Vectors(self.__op, shallow=True)
        return self.__vectors

    def arch(self):
        return self.__arch

    def gpu(self):
        return self.__gpu

    def dots(self):
        return self.__op.dots()

    def data_type(self):
        return self.__op.data_type()

    def shape(self):
        return self.__op.shape()

    def scale(self):
        return self.__scale



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

# -*- coding: utf-8 -*-
'''Simple PCA test.

Computes a number of principal components of a dataset.

If sklearn is installed, compares with sklearn.decomposition.PCA.

Usage: pca_simple <data_file> <n_components> [gpu]

data_file    : the name of the file containing data
n_components : number of principal components wanted
gpu          : run raleigh pca on GPU if this argument is present
'''

import numpy
import os
import sys
import timeit

from raleigh.interfaces.pca import pca, pca_error


narg = len(sys.argv)
if narg < 3:
    usage = 'Usage: pca_simple <data_file> <n_components> [gpu]'
    raise SystemExit(usage)
data = numpy.load(sys.argv[1])
npc = int(sys.argv[2])
m = data.shape[0]
n = data.shape[1]
if len(data.shape) > 2:
    n = numpy.prod(data.shape[1:])
    data = numpy.reshape(data, (m, n))
arch = 'cpu' if narg < 4 else 'gpu!'

numpy.random.seed(1) # make results reproducible

print('\n--- solving with raleigh pca...\n')
start = timeit.default_timer()
mean, trans, comps = pca(data, npc=npc, arch=arch)
elapsed = timeit.default_timer() - start
ncomp = comps.shape[0]
print('%d principal components computed in %.2e sec' % (ncomp, elapsed))
em, ef = pca_error(data, mean, trans, comps)
print('PCA error: max 2-norm %.0e, Frobenius norm %.0e' % (em, ef))

try:
    from sklearn.decomposition import PCA
    print('\n--- solving with sklearn PCA...')
    start = timeit.default_timer()
    skl_pca = PCA(npc)
    trans = skl_pca.fit_transform(data)
    elapsed = timeit.default_timer() - start
    comps = skl_pca.components_
    ncomp = comps.shape[0]
    print('%d principal components computed in %.2e sec' % (ncomp, elapsed))
    mean = numpy.mean(data, axis=0).reshape((1, n))
    em, ef = pca_error(data, mean, trans, comps)
    print('PCA error: max 2-norm %.0e, Frobenius norm %.0e' % (em, ef))
except:
    pass

print('done')



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""Basic core solver usage demonstration.

Usage:
    core_solver [--help | -h | options]

Options:
    -p <pro>, --problem=<pro>  eigenvalue problem type [default: std]
    -m <mat>, --matrix=<mat>   problem matrix [default: diag]
    -n <dim>, --dim=<dim>      problem size [default: 100]
    -d <dat>, --dtype=<dat>    data type (BLAS prefix s/d/c/z) [default: d]
    -l <lft>, --left=<lft>     number of left eigenvalues wanted [default: 6]
    -r <lft>, --right=<lft>    number of right eigenvalues wanted [default: 0]
    -t <tol>, --vtol=<tol>     eigenvector error tolerance [default: 1e-8]
    -b <blk>, --bsize=<blk>    block CG block size [default: -1]
    -v <vrb>, --verb=<vrb>     verbosity [default: 0]
    -P, --precond  with preconditioning
    -T, --doctest  with doctest (stay with defaults!)
"""

import numpy
import sys

from raleigh.core.solver import Options, Problem, Solver
from raleigh.core.solver import DefaultConvergenceCriteria
from raleigh.algebra.dense_cpu import Vectors, Matrix

try:
    from docopt import docopt
    have_docopt = True
except:
    have_docopt = False

if have_docopt:
    __version__ = '0.1.0'
    args = docopt(__doc__, version=__version__)
    problem = args['--problem']
    matrix = args['--matrix']
    n = int(args['--dim'])
    dt = args['--dtype']
    left = int(args['--left'])
    right = int(args['--right'])
    vec_tol = float(args['--vtol'])
    block_size = int(args['--bsize'])
    verbosity = int(args['--verb'])
    with_prec = args['--precond']
    with_doctest = args['--doctest']
else:
    print('\n=== docopt not found, using default options...\n')
    problem = 'std'
    matrix = 'diag'
    n = 100
    dt = 'd'
    left = 6
    right = 0
    vec_tol = 1e-8
    block_size = -1
    verbosity = 0
    with_prec = False
    with_doctest = True


def test():
    '''
    >>> test()
    ... # doctest: +NORMALIZE_WHITESPACE
    after 58 iterations, 6 converged eigenvalues are:
    [1. 2. 3. 4. 5. 6.]
    '''

    numpy.random.seed(1) # to debug - makes the results reproducible

    if dt == 's':
        dtype = numpy.float32
    elif dt == 'd':
        dtype = numpy.float64
    elif dt == 'c':
        dtype = numpy.complex64
    elif dt == 'z':
        dtype = numpy.complex128
    else:
        raise ValueError('data type %s not supported' % dt)

    opt = Options()
    opt.block_size = block_size
    opt.convergence_criteria = DefaultConvergenceCriteria()
    opt.convergence_criteria.set_error_tolerance('eigenvector error', vec_tol)
    opt.verbosity = verbosity

    v = Vectors(n, data_type=dtype)

    if matrix[0] == 'c':
        if dt == 's' or dt == 'd':
            raise ValueError('central differences matrix requires complex data')
        d = 1j*numpy.ones((n - 1,), dtype=dtype)
        A = numpy.diag(d, 1) - numpy.diag(d, -1)
        operatorA = Matrix(A)
    else:
        a = numpy.asarray([i + 1 for i in range(n)]).astype(dtype)
        operatorA = Matrix(numpy.diag(a))

    if problem[0] != 's':
        b = 2*numpy.ones((n,), dtype=dtype)
        operatorB = Matrix(numpy.diag(b))
    else:
        operatorB = None

    if problem[0] == 'p':
        evp = Problem(v, operatorA, operatorB, 'pro')
    else:
        evp = Problem(v, operatorA, operatorB)
    solver = Solver(evp)

    if with_prec:
        if problem[0] == 'p':
            raise ValueError('preconditioning does not work for matrix product')
        operatorP = Matrix(numpy.diag(1/a))
        solver.set_preconditioner(operatorP)

    solver.solve(v, opt, which=(left, right))
    print('after %d iterations, %d converged eigenvalues are:' \
          % (solver.iteration, v.nvec()))
    output = numpy.array_str(solver.eigenvalues)
    if output[1] == ' ':
        output = output[0] + output[2:]
    print(output)


if __name__ == '__main__':
    if with_doctest:
        import doctest
        doctest.testmod(verbose=True)
    else:
        test()



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""MKL loader/wrapper.
"""

import ctypes
import multiprocessing
import numpy
from sys import platform

from . import verbosity


def _array_ptr(array, shift=0):
    return ctypes.c_void_p(array.ctypes.data + shift)


def mkl_version(mkl):
    str_v = numpy.ndarray((256,), dtype=numpy.uint8)
    ptr_v = _array_ptr(str_v)
    len_v = ctypes.c_int(256)
    mkl.mkl_get_version_string(ptr_v, len_v)
    return str_v.tostring().decode('ascii')


if platform == 'win32':
    mkl = ctypes.CDLL('mkl_rt.dll', mode=ctypes.RTLD_GLOBAL)
else:
    mkl = ctypes.CDLL('libmkl_rt.so', mode=ctypes.RTLD_GLOBAL)
if verbosity.level > 1:
    print('Loaded %s' % mkl_version(mkl).strip())

# find the total number of threads    
num_cpus = multiprocessing.cpu_count()

# find the number of cores using MKL threading behaviour
mkl.MKL_Set_Dynamic(1)
# now mkl_get_max_threads() returns the number of cores!
num_cores = mkl.mkl_get_max_threads()

# if hyperthreading is used, increase the number of mkl threads
# to achieve slightly better performance
if num_cpus == 2*num_cores:
    num_threads = num_cpus - 1
    mkl.MKL_Set_Dynamic(0)
    mkl.MKL_Set_Num_Threads(num_threads)

if verbosity.level > 1:
    print('Using %d MKL threads' % mkl.mkl_get_max_threads())


class Cblas:
    '''MKL cblas wrapper.
    '''

    ColMajor = 102
    NoTrans = 111
    Trans = 112
    ConjTrans = 113

    def __init__(self, dt):
        if dt == numpy.float32:
            self.dsize = 4
            self.gemm = mkl.cblas_sgemm
            self.axpy = mkl.cblas_saxpy
            self.copy = mkl.cblas_scopy
            self.scal = mkl.cblas_sscal
            self.norm = mkl.cblas_snrm2
            self.norm.restype = ctypes.c_float
            self.inner = mkl.cblas_sdot
            self.inner.restype = ctypes.c_float
            self.mkl_one = ctypes.c_float(1.0)
            self.mkl_zero = ctypes.c_float(0.0)
            self.svd = mkl.LAPACKE_sgesvd
        elif dt == numpy.float64:
            self.dsize = 8
            self.gemm = mkl.cblas_dgemm
            self.axpy = mkl.cblas_daxpy
            self.copy = mkl.cblas_dcopy
            self.scal = mkl.cblas_dscal
            self.norm = mkl.cblas_dnrm2
            self.norm.restype = ctypes.c_double
            self.inner = mkl.cblas_ddot
            self.inner.restype = ctypes.c_double
            self.mkl_one = ctypes.c_double(1.0)
            self.mkl_zero = ctypes.c_double(0.0)
            self.svd = mkl.LAPACKE_dgesvd
        elif dt == numpy.complex64:
            self.dsize = 8
            self.gemm = mkl.cblas_cgemm
            self.axpy = mkl.cblas_caxpy
            self.copy = mkl.cblas_ccopy
            self.scal = mkl.cblas_cscal
            self.norm = mkl.cblas_scnrm2
            self.norm.restype = ctypes.c_float
            self.inner = mkl.cblas_cdotc_sub
            self.cmplx_val = numpy.zeros((2,), dtype=numpy.float32)
            self.cmplx_one = numpy.zeros((2,), dtype=numpy.float32)
            self.cmplx_one[0] = 1.0
            self.cmplx_zero = numpy.zeros((2,), dtype=numpy.float32)
            self.mkl_one = ctypes.c_void_p(self.cmplx_one.ctypes.data)
            self.mkl_zero = ctypes.c_void_p(self.cmplx_zero.ctypes.data)
            self.svd = mkl.LAPACKE_cgesvd
        elif dt == numpy.complex128:
            self.dsize = 16
            self.gemm = mkl.cblas_zgemm
            self.axpy = mkl.cblas_zaxpy
            self.copy = mkl.cblas_zcopy
            self.scal = mkl.cblas_zscal
            self.norm = mkl.cblas_dznrm2
            self.norm.restype = ctypes.c_double
            self.inner = mkl.cblas_zdotc_sub
            self.cmplx_val = numpy.zeros((2,), dtype=numpy.float64)
            self.cmplx_one = numpy.zeros((2,), dtype=numpy.float64)
            self.cmplx_one[0] = 1.0
            self.cmplx_zero = numpy.zeros((2,), dtype=numpy.float64)
            self.mkl_one = ctypes.c_void_p(self.cmplx_one.ctypes.data)
            self.mkl_zero = ctypes.c_void_p(self.cmplx_zero.ctypes.data)
            self.svd = mkl.LAPACKE_zgesvd
        else:
            raise ValueError('data type %s not supported' % repr(dt))
        self.__s = 'S'
        self.__o = 'O'
        self.small = ctypes.c_char(self.__s.encode('utf-8'))
        self.ovwrt = ctypes.c_char(self.__o.encode('utf-8'))


class SparseSymmetricMatrix:
    '''MKL csr matrix wrapper.
    '''

    def __init__(self, a, ia, ja):
        self.__a = a
        self.__ib = ia
        self.__ie = ia[1:]
        self.__ja = ja
        self.__n = ctypes.c_int(ia.shape[0] - 1)
        self.__dtype = a.dtype
        self.__oz = numpy.array([1.0, 0.0]).astype(a.dtype)
        self.__one = _array_ptr(self.__oz)
        self.__zero = _array_ptr(self.__oz[1:])
        self.__complex = False
        if a.dtype == numpy.float32:
            self.__csrmm = mkl.mkl_scsrmm
            self.__csrsymv = mkl.mkl_scsrsymv
            self.__descr = 'SUNF  '
        elif a.dtype == numpy.float64:
            self.__csrmm = mkl.mkl_dcsrmm
            self.__csrsymv = mkl.mkl_dcsrsymv
            self.__descr = 'SUNF  '
        elif a.dtype == numpy.complex64:
            self.__csrmm = mkl.mkl_ccsrmm
            self.__csrsymv = mkl.mkl_ccsrsymv
            self.__descr = 'HUNF  '
            self.__complex = True
        elif a.dtype == numpy.complex128:
            self.__csrmm = mkl.mkl_zcsrmm
            self.__csrsymv = mkl.mkl_zcsrsymv
            self.__descr = 'HUNF  '
            self.__complex = True
        else:
            raise ValueError('unsupported data type')
        self.__csrmm.restype = None
        self.__uplo = 'U'
        self.__trans = 'N'
        self.__u = ctypes.c_char_p(self.__uplo.encode('utf-8'))
        self.__t = ctypes.c_char_p(self.__trans.encode('utf-8'))
        self.__d = ctypes.c_char_p(self.__descr.encode('utf-8'))

    def dot(self, x, y):
        ptr_x = _array_ptr(x)
        ptr_y = _array_ptr(y)
        ptr_a = _array_ptr(self.__a)
        ptr_ib = _array_ptr(self.__ib)
        ptr_ie = _array_ptr(self.__ie)
        ptr_ja = _array_ptr(self.__ja)
        s = x.shape
        if (len(s) == 1 or s[0] == 1) and not self.__complex:
            #print('using *csrsymv...')
            if len(s) == 1:
                n = s[0]
            else:
                n = numpy.prod(s[1:])
            n = ctypes.c_int(n)
            self.__csrsymv(self.__u, ctypes.byref(n), \
                           ptr_a, ptr_ib, ptr_ja, ptr_x, ptr_y)
            return
        #print('using *csrmm...')
        if len(s) > 1:
            m, n = s[0], int(numpy.prod(s[1:]))
        else:
            m = 1
            n = self.__n.value
        mn = numpy.array([m, n])
        ptr_m = _array_ptr(mn)
        ptr_n = _array_ptr(mn[1:])
        n = self.__n
        self.__csrmm(self.__t, ctypes.byref(n), ptr_m, ctypes.byref(n), \
            self.__one, self.__d, ptr_a, ptr_ja, ptr_ib, ptr_ie, \
            ptr_x, ptr_n, self.__zero, ptr_y, ptr_n)


class ILUT:
    '''MKL dscrilut wrapper.
    '''

    def __init__(self, a, ia, ja):
        self.__a = a
        self.__ia = ia
        self.__ja = ja
        n = ia.shape[0] - 1
        nnz = ia[n] - ia[0]
        self.__rows = n
        self.__nnz = nnz//n
        self.__ipar = numpy.zeros((128,), dtype=numpy.int32)
        self.__dpar = numpy.zeros((128,))
        self.__w = numpy.zeros((n,))
        self.__ipar[0] = n
        self.__b = None
        self.__ib = None
        self.__jb = None
        self.__u = 'U'
        self.__l = 'L'
        self.__n = 'N'
        self.__U = ctypes.c_char_p(self.__u.encode('utf-8'))
        self.__L = ctypes.c_char_p(self.__l.encode('utf-8'))
        self.__N = ctypes.c_char_p(self.__n.encode('utf-8'))

    def factorize(self, tol=1e-6, max_fill_rel=1):
        max_fill_abs = min(self.__rows - 1, self.__nnz * max_fill_rel)
        self.__ipar[30] = 1
        self.__dpar[30] = tol
        n = self.__rows
        nnz = (2*max_fill_abs + 1)*n
        #print(n, nnz)
        self.__b = numpy.ndarray((nnz,)) 
        self.__ib = numpy.ndarray((n + 1,)) 
        self.__jb = numpy.ndarray((nnz,), dtype=numpy.int32)
        n_c = ctypes.c_int(n)
        mf_c = ctypes.c_int(max_fill_abs)
        ierr_c = ctypes.c_int()
        tol_c = ctypes.c_double(tol)
        ptr_a = _array_ptr(self.__a)
        ptr_ia = _array_ptr(self.__ia)
        ptr_ja = _array_ptr(self.__ja)
        ptr_b = _array_ptr(self.__b)
        ptr_ib = _array_ptr(self.__ib)
        ptr_jb = _array_ptr(self.__jb)
        ptr_ipar = _array_ptr(self.__ipar)
        ptr_dpar = _array_ptr(self.__dpar)
        mkl.dcsrilut(ctypes.byref(n_c), \
                     ptr_a, ptr_ia, ptr_ja, ptr_b, ptr_ib, ptr_jb, \
                     ctypes.byref(tol_c), ctypes.byref(mf_c), \
                     ptr_ipar, ptr_dpar, ctypes.byref(ierr_c))
        if ierr_c.value != 0: print(ierr_c.value)

    def solve(self, f, x):
        m, n = f.shape
        n = ctypes.c_int(self.__rows)
        #print(n)
        ptr_b = _array_ptr(self.__b)
        ptr_ib = _array_ptr(self.__ib)
        ptr_jb = _array_ptr(self.__jb)
        ptr_w = _array_ptr(self.__w)
        for i in range(m):
            ptr_f = _array_ptr(f[i,:])
            ptr_x = _array_ptr(x[i,:])
            mkl.mkl_dcsrtrsv(self.__L, self.__N, self.__U, ctypes.byref(n), \
                             ptr_b, ptr_ib, ptr_jb, ptr_f, ptr_w)
            mkl.mkl_dcsrtrsv(self.__U, self.__N, self.__N, ctypes.byref(n), \
                             ptr_b, ptr_ib, ptr_jb, ptr_w, ptr_x)


class ParDiSo:
    '''MKL pardiso wrapper.
    '''

    def __init__(self, dtype=numpy.float64, pos_def=False):
        self.__pardiso = mkl.pardiso
        self.__pt = numpy.ndarray((64,), dtype=numpy.int64)
        self.__iparm = numpy.ndarray((64,), dtype=numpy.int32)
        self.__handle = _array_ptr(self.__pt)
        if (dtype == numpy.float32 or dtype == numpy.float64):
            m = 2
            self.__real = True
        elif (dtype == numpy.complex64 or dtype == numpy.complex128):
            m = 4
            self.__real = False
        else:
            raise ValueError('ParDiSo constructor: wrong data type')
        if not pos_def:
            m = -m
        #print('matrix type: %d' % m)
        mtype = ctypes.c_int(m)
        self.__dtype = dtype
        self.__mtype = mtype
        self.__a = None
        self.__ia = None
        self.__ja = None
        self.__perm = None
        self.__dummy = numpy.ndarray((1,))
        self.__ptr = ctypes.c_void_p(self.__dummy.ctypes.data)
        self.__ptr_iparm = _array_ptr(self.__iparm)
        mkl.pardisoinit(self.__handle, ctypes.byref(mtype), self.__ptr_iparm)
        self.__iparm[4] = 2
        if dtype == numpy.float32 or dtype == numpy.complex64:
            self.__iparm[27] = 1

    def __del__(self):
        step = ctypes.c_int(-1)
        maxf = ctypes.c_int(1)
        mnum = ctypes.c_int(1)
        n = ctypes.c_int(1)
        m = ctypes.c_int(1)
        ptr = self.__ptr
        verb = ctypes.c_int(0)
        err = ctypes.c_long(0)
        self.__pardiso(self.__handle, \
            ctypes.byref(maxf), ctypes.byref(mnum), \
            ctypes.byref(self.__mtype), \
            ctypes.byref(step), ctypes.byref(n), ptr, ptr, ptr, \
            ptr, ctypes.byref(m), self.__ptr_iparm, ctypes.byref(verb), \
            ptr, ptr, ctypes.byref(err))
        if err.value != 0: print(err.value)

    def handle(self):
        return self.__handle

    def iparm(self):
        return self.__iparm

    def perm(self):
        return self.__perm

    def analyse(self, a, ia, ja):
        self.__a = a
        self.__ia = ia
        self.__ja = ja
        step = ctypes.c_int(11)
        maxf = ctypes.c_int(1)
        mnum = ctypes.c_int(1)
        rows = ia.shape[0] - 1
        n = ctypes.c_int(rows)
        m = ctypes.c_int(1)
        ptr_a = _array_ptr(self.__a)
        ptr_ia = _array_ptr(self.__ia)
        ptr_ja = _array_ptr(self.__ja)
        self.__perm = numpy.ndarray((rows,), dtype=numpy.int32)
        ptr_perm = _array_ptr(self.__perm)
        ptr_iparm = _array_ptr(self.__iparm)
        verb = ctypes.c_int(0)
        err = ctypes.c_long(0)
        self.__pardiso(self.__handle, ctypes.byref(maxf), ctypes.byref(mnum), \
                    ctypes.byref(self.__mtype), \
                    ctypes.byref(step), ctypes.byref(n), ptr_a, ptr_ia, ptr_ja, \
                    ptr_perm, ctypes.byref(m), ptr_iparm, ctypes.byref(verb), \
                    self.__ptr, self.__ptr, ctypes.byref(err))
        if err.value != 0: print(err.value)

    def factorize(self):
        step = ctypes.c_int(22)
        maxf = ctypes.c_int(1)
        mnum = ctypes.c_int(1)
        rows = self.__ia.shape[0] - 1
        n = ctypes.c_int(rows)
        m = ctypes.c_int(1)
        ptr_a = _array_ptr(self.__a)
        ptr_ia = _array_ptr(self.__ia)
        ptr_ja = _array_ptr(self.__ja)
        ptr_perm = _array_ptr(self.__perm)
        ptr_iparm = _array_ptr(self.__iparm)
        verb = ctypes.c_int(0)
        err = ctypes.c_long(0)
        self.__pardiso(self.__handle, ctypes.byref(maxf), ctypes.byref(mnum), \
                    ctypes.byref(self.__mtype), \
                    ctypes.byref(step), ctypes.byref(n), ptr_a, ptr_ia, ptr_ja, \
                    ptr_perm, ctypes.byref(m), ptr_iparm, ctypes.byref(verb), \
                    self.__ptr, self.__ptr, ctypes.byref(err))
        if err.value != 0: print(err.value)

    def solve(self, b, x, part = None):
        if len(b.shape) > 1:
            nrhs = b.shape[0]
        else:
            nrhs = 1
        if part == 'f':
            step = ctypes.c_int(331)
        elif part == 'd':
            step = ctypes.c_int(332)
        elif part == 'b':
            step = ctypes.c_int(333)
        else:
            step = ctypes.c_int(33)
        maxf = ctypes.c_int(1)
        mnum = ctypes.c_int(1)
        rows = self.__ia.shape[0] - 1
        n = ctypes.c_int(rows)
        m = ctypes.c_int(nrhs)
        ptr_a = _array_ptr(self.__a)
        ptr_ia = _array_ptr(self.__ia)
        ptr_ja = _array_ptr(self.__ja)
        ptr_perm = _array_ptr(self.__perm)
        ptr_iparm = _array_ptr(self.__iparm)
        verb = ctypes.c_int(0)
        err = ctypes.c_long(0)
        ptr_b = _array_ptr(b)
        ptr_x = _array_ptr(x)
        self.__pardiso(self.__handle, ctypes.byref(maxf), ctypes.byref(mnum), \
                    ctypes.byref(self.__mtype), \
                    ctypes.byref(step), ctypes.byref(n), ptr_a, ptr_ia, ptr_ja, \
                    ptr_perm, ctypes.byref(m), ptr_iparm, ctypes.byref(verb), \
                    ptr_b, ptr_x, ctypes.byref(err))
        if err.value != 0: print(err.value)

    def diag(self):
        rows = self.__ia.shape[0] - 1
        perm = self.__perm - 1
        f = numpy.zeros((4, rows), dtype=self.__dtype)
        w = numpy.zeros((4, rows), dtype=self.__dtype)
        for i in range(rows):
            j = i%4
            f[j, perm[i]] = 1
        self.solve(f, w, part = 'd')
        w = w[:,perm]
        d = numpy.zeros((2, rows), dtype=self.__dtype)
        i = 0
        while i < rows:
            if i > 0:
                d[1, i - 1] = w[0, i - 1]
            d[0, i] = w[0, i]
            d[1, i] = w[1, i]
            if i == rows - 1:
                break
            i += 1
            d[0, i] = w[1, i]
            d[1, i] = w[2, i]
            if i == rows - 1:
                break
            i += 1
            d[0, i] = w[2, i]
            d[1, i] = w[3, i]
            if i == rows - 1:
                break
            i += 1
            d[0, i] = w[3, i]
            i += 1
        return d

    def inertia(self):
        if self.__real:
            return self.__iparm[22], self.__iparm[21]
        rows = self.__ia.shape[0] - 1
        diags = self.diag()
        diag = numpy.real(diags[0, :])
        offd = diags[1, :]
        nneg = 0
        npos = 0
        i = 0
        while i < rows:
            a = diag[i]
            c = offd[i]
            if c == 0 or i == rows - 1:
                if a < 0:
                    nneg += 1
                elif a > 0:
                    npos += 1
                i += 1
            else:
                b = diag[i + 1]
                det = a*b - abs(c)**2
                s = a + b
                if det == 0:
                    if s < 0:
                        nneg += 1
                    elif s > 0:
                        npos += 1
                elif det < 0:
                    nneg += 1
                    npos += 1
                elif s < 0:
                    nneg += 2
                i += 2
        return nneg, npos




'''Example scripts.

core_solver.py:
    Basic core solver usage demonstration.

sparse_evp.py:
    Computes several eigenvalues and eigenvectors of a real symmetric matrix.

truncated_svd.py:
    Truncated SVD and PCA for a randomly generated matrix.
'''



'''An example of RALEIGH application to computing eigenimages (Principal
   Components of 2D images sets).

convert_lfw.py:
    Reads images from an lfw-style folder (see http://vis-www.cs.umass.edu/lfw),
    optionally processes them and saves in .npy file.
    === MUST BE RUN FIRST ===

compute_eigenimages.py:
    Computes Principal Components for a set of 2D images in small portions until
    stopped by user's entering 'n' in answer to 'more?' or the error of PCA
    approximation falls below the required tolerance.

show_errors.py:
    Compares images with their PCA approximations computed by the script
    compute_eigenimages.py.
'''



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""CUBLAS loader/wrapper.
"""

import ctypes
import glob
import numpy
from sys import platform

from . import verbosity


POINTER = ctypes.POINTER


try:
    if platform == 'win32':
        from .cuda_wrap import cuda_path
        cublas_dll = glob.glob(cuda_path + '/cublas64*')[0]
        cublas = ctypes.CDLL(cublas_dll, mode=ctypes.RTLD_GLOBAL)
        if verbosity.level > 0:
            print('loaded %s' % cublas_dll)
    else:
        cublas = ctypes.CDLL('libcublas.so', mode=ctypes.RTLD_GLOBAL)

    cublasCreate = cublas.cublasCreate_v2
    cublasCreate.argtypes = [POINTER(POINTER(ctypes.c_ubyte))]
    cublasCreate.restype = ctypes.c_int

    cublasDestroy = cublas.cublasDestroy_v2
    cublasDestroy.restype = ctypes.c_int

    cublasVersion = cublas.cublasGetVersion_v2
    cublasVersion.restype = ctypes.c_int

    cublas_handle = POINTER(ctypes.c_ubyte)()
    err = cublasCreate(ctypes.byref(cublas_handle))
    v = ctypes.c_int()
    err = cublasVersion(cublas_handle, ctypes.byref(v))
    version = v.value
    if verbosity.level > 0:
        print('CUBLAS version: %d' % version)
    cublasDestroy(cublas_handle)

except:
    if verbosity.level > 0:
        print('CUBLAS not found, switching to cpu...')
    raise RuntimeError('CUBLAS not found')

##try:
##    if platform == 'win32':
##        from .cuda_wrap import cuda_path
##        cusolver_dll = glob.glob(cuda_path + '/cusolver64*')[0]
##        cusolver = ctypes.CDLL(cusolver_dll, mode=ctypes.RTLD_GLOBAL)
##        if verbosity.level > 0:
##            print('loaded %s' % cusolver_dll)
##    else:
##        cusolver = ctypes.CDLL('libcusolver.so', mode=ctypes.RTLD_GLOBAL)
##
##    cusolverCreate = cusolver.cusolverDnCreate
##    cusolverCreate.argtypes = [POINTER(POINTER(ctypes.c_ubyte))]
##    cusolverCreate.restype = ctypes.c_int
##
##    cusolverDestroy = cusolver.cusolverDnDestroy
##    cusolverDestroy.restype = ctypes.c_int
##
##except:
##    if verbosity.level > -1:
##        print('cuSOLVER library not found')
##    raise RuntimeError('CUSOLVER not found')


class Cublas:
    '''CUBLAS wrapper.
    '''

    CtypesPtr = POINTER(ctypes.c_ubyte)
    NoTrans = 0
    Trans = 1
    ConjTrans = 2
    HostMode = ctypes.c_int(0)
    DeviceMode = ctypes.c_int(1)

    def __init__(self, dt):
#        print('creating cublas...')
        self.handle = Cublas.CtypesPtr()
        self.cusolver_handle = Cublas.CtypesPtr()
        err = cublasCreate(ctypes.byref(self.handle))
        if err != 0:
            raise RuntimeError('cublasCreate failure, cuda error %d' % err)
        self.__cublas = cublas
##        err = cusolverCreate(ctypes.byref(self.cusolver_handle))
##        if err != 0:
##            raise RuntimeError('cusolverCreate failure, cuda error %d' % err)
        self.getPointerMode = cublas.cublasGetPointerMode_v2
        self.getPointerMode.restype = ctypes.c_int
        self.setPointerMode = cublas.cublasSetPointerMode_v2
        self.setPointerMode.restype = ctypes.c_int
        self.setStream = cublas.cublasSetStream_v2
        self.setStream.restype = ctypes.c_int
        if dt == numpy.float32:
            self.dsize = 4
            self.copy = cublas.cublasScopy_v2
            self.axpy = cublas.cublasSaxpy_v2
            self.scal = cublas.cublasSscal_v2
            self.norm = cublas.cublasSnrm2_v2
            self.dot = cublas.cublasSdot_v2
            self.gemm = cublas.cublasSgemm_v2
            self.gemm_batched = cublas.cublasSgemmBatched
##            self.svd_buff_size = cusolver.cusolverDnSgesvd_bufferSize
##            self.svd = cusolver.cusolverDnSgesvd
        elif dt == numpy.float64:
            self.dsize = 8
            self.copy = cublas.cublasDcopy_v2
            self.axpy = cublas.cublasDaxpy_v2
            self.scal = cublas.cublasDscal_v2
            self.norm = cublas.cublasDnrm2_v2
            self.dot = cublas.cublasDdot_v2
            self.gemm = cublas.cublasDgemm_v2
            self.gemm_batched = cublas.cublasDgemmBatched
##            self.svd_buff_size = cusolver.cusolverDnDgesvd_bufferSize
##            self.svd = cusolver.cusolverDnDgesvd
        elif dt == numpy.complex64:
            self.dsize = 8
            self.copy = cublas.cublasCcopy_v2
            self.axpy = cublas.cublasCaxpy_v2
            self.scal = cublas.cublasCscal_v2
            self.fscal = cublas.cublasSscal_v2
            self.fscal.restype = ctypes.c_int
            self.norm = cublas.cublasScnrm2_v2
            self.dot = cublas.cublasCdotc_v2
            self.gemm = cublas.cublasCgemm_v2
            self.gemm_batched = cublas.cublasCgemmBatched
##            self.svd_buff_size = cusolver.cusolverDnCgesvd_bufferSize
##            self.svd = cusolver.cusolverDnCgesvd
        elif dt == numpy.complex128:
            self.dsize = 16
            self.copy = cublas.cublasZcopy_v2
            self.axpy = cublas.cublasZaxpy_v2
            self.scal = cublas.cublasZscal_v2
            self.fscal = cublas.cublasDscal_v2
            self.fscal.restype = ctypes.c_int
            self.norm = cublas.cublasDznrm2_v2
            self.dot = cublas.cublasZdotc_v2
            self.gemm = cublas.cublasZgemm_v2
            self.gemm_batched = cublas.cublasZgemmBatched
##            self.svd_buff_size = cusolver.cusolverDnZgesvd_bufferSize
##            self.svd = cusolver.cusolverDnZgesvd
        else:
            raise ValueError('data type %s not supported' % repr(dt))
        self.copy.restype = ctypes.c_int
        self.axpy.restype = ctypes.c_int
        self.scal.restype = ctypes.c_int
        self.norm.restype = ctypes.c_int
        self.dot.restype = ctypes.c_int
        self.gemm.restype = ctypes.c_int
        self.gemm_batched.restype = ctypes.c_int
##        self.svd_buff_size.restype = ctypes.c_int
##        self.svd.argtypes = [POINTER(ctypes.c_ubyte), \
##            ctypes.c_char, ctypes.c_char, ctypes.c_int, ctypes.c_int, \
##            POINTER(ctypes.c_ubyte), ctypes.c_int, POINTER(ctypes.c_ubyte), \
##            POINTER(ctypes.c_ubyte), ctypes.c_int, \
##            POINTER(ctypes.c_ubyte), ctypes.c_int, \
##            POINTER(ctypes.c_ubyte), ctypes.c_int, \
##            POINTER(ctypes.c_ubyte), POINTER(ctypes.c_ubyte)]
##        self.svd.restype = ctypes.c_int
        self.all = ctypes.c_char('A'.encode('utf-8'))
        self.small = ctypes.c_char('S'.encode('utf-8'))
        self.ovwrt = ctypes.c_char('O'.encode('utf-8'))

    def __del__(self):
#        print('destroying cublas...')
        self.__cublas.cublasDestroy_v2(self.handle)
##        cusolverDestroy(self.cusolver_handle)



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""MKL cblas implementation of RALEIGH dense algebra.
"""

import ctypes
import numpy

from .mkl_wrap import Cblas
from .dense_ndarray import NDArrayVectors, NDArrayMatrix


class Vectors(NDArrayVectors):
    '''MKL cblas implementation of ndarray-based Vectors type.
    '''

    '''========== Methods required by RALEIGH core solver ==========
    '''

    def new_vectors(self, arg=0, dim=None):
        if isinstance(arg, numpy.ndarray):
            return Vectors(arg)
        nv = arg
        if dim is None:
            dim = self.dimension()
        return Vectors(dim, nv, self.data_type())

    def clone(self):
        return Vectors(self)

    def copy(self, other, ind=None):
        i, n = self.selected()
        j, m = other.selected()
        vdim = self.dimension()
        mkl_inc = ctypes.c_int(1)
        vsize = self.__cblas.dsize * vdim
        if ind is None:
            mkl_n = ctypes.c_int(n*vdim)
            ptr_u = _array_ptr(self.all_data(), i*vsize)
            ptr_v = _array_ptr(other.all_data(), j*vsize)
            self.__cblas.copy(mkl_n, ptr_u, mkl_inc, ptr_v, mkl_inc)
        else:
            mkl_n = ctypes.c_int(vdim)
            l = len(ind)
            for k in range(l):
                ptr_u = _array_ptr(self.all_data(), int(ind[k])*vsize)
                ptr_v = _array_ptr(other.all_data(), (j + k)*vsize)
                self.__cblas.copy(mkl_n, ptr_u, mkl_inc, ptr_v, mkl_inc)

    def scale(self, s, multiply=False):
        f, n = self.selected()
        vdim = self.dimension()
        mkl_n = ctypes.c_int(vdim)
        mkl_inc = ctypes.c_int(1)
        vsize = self.__cblas.dsize * vdim
        if multiply:
            for i in range(n):
                ptr_u = _array_ptr(self.all_data(), (f + i)*vsize)
                mkl_s = self.__to_float(s[i])
                self.__cblas.scal(mkl_n, mkl_s, ptr_u, mkl_inc)
        else:
            for i in range(n):
                if s[i] != 0.0:
                    ptr_u = _array_ptr(self.all_data(), (f + i)*vsize)
                    mkl_s = self.__to_float(1.0/s[i])
                    self.__cblas.scal(mkl_n, mkl_s, ptr_u, mkl_inc)

    def dots(self, other, transp=False):
        if transp:
            u = self.data()
            v = other.data()
            n = self.dimension()
            w = numpy.ndarray((n,), dtype=self.data_type())
            if other.is_complex():
                for i in range(n):
                    w[i] = numpy.dot(v[:, i].conj(), u[:, i])
            else:
                for i in range(n):
                    w[i] = numpy.dot(v[:, i], u[:, i])
            return w
        iu = self.selected()[0]
        iv = other.selected()[0]
        vdim = self.dimension()
        dsize = self.__cblas.dsize
        vsize = dsize * vdim
        n = self.nvec()
        mkl_n = ctypes.c_int(vdim)
        mkl_inc = ctypes.c_int(1)
        w = numpy.ndarray((n,), dtype=self.data_type())
        for i in range(n):
            ptr_u = _array_ptr(self.all_data(), (iu + i)*vsize)
            ptr_v = _array_ptr(other.all_data(), (iv + i)*vsize)
            if self.is_complex():
                s = self.__complex()
                self.__cblas.inner \
                    (mkl_n, ptr_v, mkl_inc, ptr_u, mkl_inc, s)
                w[i] = s[0] + 1j * s[1]
            else:
                w[i] = self.__cblas.inner \
                    (mkl_n, ptr_v, mkl_inc, ptr_u, mkl_inc)
        return w

    def dot(self, other):
        n = self.dimension()
        m = self.nvec()
        k = other.nvec()
        q = numpy.ndarray((k, m), dtype=self.data_type())
        mkl_n = ctypes.c_int(n)
        mkl_m = ctypes.c_int(m)
        mkl_k = ctypes.c_int(k)
        ptr_u = _array_ptr(other.data())
        ptr_v = _array_ptr(self.data())
        ptr_q = ctypes.c_void_p(q.ctypes.data)
        if self.is_complex():
            Trans = Cblas.ConjTrans
        else:
            Trans = Cblas.Trans
        self.__cblas.gemm(Cblas.ColMajor, Trans, Cblas.NoTrans, \
            mkl_m, mkl_k, mkl_n, \
            self.__cblas.mkl_one, ptr_v, mkl_n, ptr_u, mkl_n, \
            self.__cblas.mkl_zero, ptr_q, mkl_m)
        return _conjugate(q)

    def multiply(self, q, output):
        f, s = output.selected();
        m = q.shape[1]
        n = self.dimension()
        mkl_n = ctypes.c_int(n)
        mkl_m = ctypes.c_int(m)
        mkl_k = ctypes.c_int(self.nvec())
        ptr_u = _array_ptr(output.data())
        ptr_v = _array_ptr(self.data())
        ptr_q = ctypes.c_void_p(q.ctypes.data)
        if q.flags['C_CONTIGUOUS']:
            Trans = Cblas.Trans
            ldq = mkl_m
        elif q.flags['F_CONTIGUOUS']:
            Trans = Cblas.NoTrans
            ldq = mkl_k
        else:
            #print('using non-optimized dot')
            output.data()[:,:] = numpy.dot(q.T, self.data())
            return
        self.__cblas.gemm(Cblas.ColMajor, Cblas.NoTrans, Trans, \
            mkl_n, mkl_m, mkl_k, \
            self.__cblas.mkl_one, ptr_v, mkl_n, ptr_q, ldq, \
            self.__cblas.mkl_zero, ptr_u, mkl_n)

    def add(self, other, s, q=None):
        f, m = self.selected();
        n = self.dimension()
        m = other.nvec()
        mkl_n = ctypes.c_int(n)
        mkl_m = ctypes.c_int(m)
        vsize = self.__cblas.dsize * n
        ptr_u = _array_ptr(other.data())
        ptr_v = _array_ptr(self.data())
        if numpy.isscalar(s):
            mkl_s = self.__to_float(s)
            if q is None:
                mkl_nm = ctypes.c_int(n*m)
                mkl_inc = ctypes.c_int(1)
                self.__cblas.axpy \
                    (mkl_nm, mkl_s, ptr_u, mkl_inc, ptr_v, mkl_inc)
            else:
                mkl_k = ctypes.c_int(q.shape[1])
                ptr_q = ctypes.c_void_p(q.ctypes.data)
                if q.flags['C_CONTIGUOUS']:
                    Trans = Cblas.Trans
                    ldq = mkl_k
                elif q.flags['F_CONTIGUOUS']:
                    Trans = Cblas.NoTrans
                    ldq = mkl_m
                else:
                    #print('using non-optimized dot')
                    self.data()[:,:] += s*numpy.dot(q.T, other.data())
                    return
                self.__cblas.gemm(Cblas.ColMajor, Cblas.NoTrans, Trans, \
                    mkl_n, mkl_k, mkl_m, \
                    mkl_s, ptr_u, mkl_n, ptr_q, ldq, \
                    self.__cblas.mkl_one, ptr_v, mkl_n)
        else:
            for i in range(m):
                ptr_u = _array_ptr(other.data(), i*vsize)
                ptr_v = _array_ptr(self.data(), i*vsize)
                mkl_inc = ctypes.c_int(1)
                mkl_s = self.__to_float(s[i])
                self.__cblas.axpy \
                    (mkl_n, mkl_s, ptr_u, mkl_inc, ptr_v, mkl_inc)

    '''========== Other methods ====================================
    '''

    def __init__(self, arg, nvec=0, data_type=None, shallow=False):
        super(Vectors, self).__init__(arg, nvec, data_type, shallow)
        dt = self.data_type()
        self.__cblas = Cblas(dt)

    def __complex(self):
        dt = self.data_type()
        if dt == numpy.complex64:
            floats = ctypes.c_float * 2
        elif dt == numpy.complex128:
            floats = ctypes.c_double * 2
        else:
            raise ValueError('wrong data type %s passed to __complex' % repr(dt))
        return floats()

    def __to_float(self, v):
        dt = self.data_type()
        if dt == numpy.float32:
            s = ctypes.c_float(v)
        elif dt == numpy.float64:
            s = ctypes.c_double(v)
        elif dt == numpy.complex64 or dt == numpy.complex128:
            s = self.__complex()
            s[0] = v.real
            s[1] = v.imag
        else:
            raise ValueError('data type %s not supported' % repr(dt))
        return s

    def cblas(self):
        return self.__cblas

    def reference(self):
        v = Vectors(self, shallow=True)
        return v

    def orthogonalize(self, other):
        m = self.nvec()
        n = self.dimension()
        k = other.nvec()
        q = self.new_vectors(k, m)
        c_n = ctypes.c_int(n)
        c_m = ctypes.c_int(m)
        c_k = ctypes.c_int(k)
        ptr_u = _array_ptr(other.data())
        ptr_v = _array_ptr(self.data())
        ptr_q = _array_ptr(q.data())
        if self.is_complex():
            Trans = Cblas.ConjTrans
        else:
            Trans = Cblas.Trans
        zero = self.__to_float(0.0)
        pl_one = self.__to_float(1.0)
        mn_one = self.__to_float(-1.0)
        self.__cblas.gemm(Cblas.ColMajor, Trans, Cblas.NoTrans, \
            c_m, c_k, c_n, pl_one, ptr_v, c_n, ptr_u, c_n, zero, ptr_q, c_m)
        self.__cblas.gemm(Cblas.ColMajor, Cblas.NoTrans, Trans, \
            c_n, c_m, c_k, mn_one, ptr_u, c_n, ptr_q, c_m, pl_one, ptr_v, c_n)
        return q

    def svd(self):
        m = self.nvec()
        n = self.dimension()
        c_n = ctypes.c_int(n)
        c_m = ctypes.c_int(m)
        v = numpy.ndarray((m, m), dtype=self.data_type())
        dtype = self.data_type()
        if dtype == numpy.complex64:
            dtype = numpy.float32
        elif dtype == numpy.complex128:
            dtype = numpy.float64            
        sigma = numpy.ndarray((m,), dtype=dtype)
        sd = numpy.ndarray((m,), dtype=self.data_type())
        ptr_u = _array_ptr(self.data())
        ptr_v = ctypes.c_void_p(v.ctypes.data)
        ptr_s = ctypes.c_void_p(sigma.ctypes.data)
        ptr_sd = ctypes.c_void_p(sd.ctypes.data)
        jobu = self.__cblas.ovwrt
        jobv = self.__cblas.small
        self.__cblas.svd(Cblas.ColMajor, jobu, jobv, c_n, c_m, \
            ptr_u, c_n, ptr_s, ptr_u, c_n, ptr_v, c_m, ptr_sd)
        return sigma, v.T

    def apply(self, A, output, transp=False):
        a = A.data()
        if transp:
            is_complex = (a.dtype.kind == 'c')
            if is_complex:
                numpy.conj(self.data(), out=self.data())
            self.__apply(a.T, output)
            if is_complex:
                numpy.conj(output.data(), out=output.data())
                numpy.conj(self.data(), out=self.data())
        else:
            self.__apply(a, output)

    def __apply(self, q, output):
        f, s = output.selected();
        m = q.shape[0]
        n = self.dimension()
        k = self.nvec()
        mkl_n = ctypes.c_int(n)
        mkl_m = ctypes.c_int(m)
        mkl_k = ctypes.c_int(k)
        if q.flags['C_CONTIGUOUS']:
            Trans = Cblas.Trans
            ldq = mkl_n
        elif q.flags['F_CONTIGUOUS']:
            Trans = Cblas.NoTrans
            ldq = mkl_m
        else:
            #print('using non-optimized dot')
            output.data()[:,:] = numpy.dot(self.data(), q.T)
            return
        ptr_u = _array_ptr(output.data())
        ptr_v = _array_ptr(self.data())
        ptr_q = ctypes.c_void_p(q.ctypes.data)
        self.__cblas.gemm(Cblas.ColMajor, Trans, Cblas.NoTrans, \
            mkl_m, mkl_k, mkl_n, \
            self.__cblas.mkl_one, ptr_q, ldq, ptr_v, mkl_n, \
            self.__cblas.mkl_zero, ptr_u, mkl_m)


class Matrix(NDArrayMatrix):

    def apply(self, x, y, transp=False):
        if transp:
            is_complex = self.is_complex()
            if is_complex:
                numpy.conj(x.data(), out=x.data())
            self.__apply(x, y, transp)
            if is_complex:
                numpy.conj(x.data(), out=x.data())
                numpy.conj(y.data(), out=y.data())
        else:
            self.__apply(x, y)

    def __apply(self, x, y, transp=False):
        if transp:
            q = self.data().T
        else:
            q = self.data()
        m = q.shape[0]
        n = x.dimension()
        k = x.nvec()
        mkl_n = ctypes.c_int(n)
        mkl_m = ctypes.c_int(m)
        mkl_k = ctypes.c_int(k)
        if q.flags['C_CONTIGUOUS']:
            Trans = Cblas.Trans
            ldq = mkl_n
        elif q.flags['F_CONTIGUOUS']:
            Trans = Cblas.NoTrans
            ldq = mkl_m
        else:
            #print('using non-optimized dot')
            y.data()[:,:] = numpy.dot(x.data(), q.T)
            return
        ptr_u = _array_ptr(y.data())
        ptr_v = _array_ptr(x.data())
        ptr_q = ctypes.c_void_p(q.ctypes.data)
        x.cblas().gemm(Cblas.ColMajor, Trans, Cblas.NoTrans, \
            mkl_m, mkl_k, mkl_n, \
            x.cblas().mkl_one, ptr_q, ldq, ptr_v, mkl_n, \
            x.cblas().mkl_zero, ptr_u, mkl_m)

    def dots(self):
        v = Vectors(self, shallow=True)
        return v.dots(v)

    def new_vectors(self, dim=None, nv=0):
        if dim is None:
            dim = self.shape()[1]
        return Vectors(dim, nv, self.data_type())

def _array_ptr(array, shift=0):
    return ctypes.c_void_p(array.ctypes.data + shift)


def _conjugate(a):
    if a.dtype.kind == 'c':
        return a.conj()
    else:
        return a



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

'''Principal Component Analysis of 2D images.

Computes Principal Components for a set of 2D images in small portions until
stopped by user's entering 'n' in answer to 'more?' or the error of PCA
approximation falls below the tolerance.

Note on Linear Algebra style terminology adopted here:
Due to a specific nature of this PCA application, principal components are
referred to as 'eigenimages'. Each image from the original data set is
approximated by a linear combination of eigenimages, hence we refer to the
coefficients of these linear combinations as 'coordinates' (of images in
the orthonormal basis of eigenimages) - in some other PCA software these
are referred to as 'reduced features data'.

Usage:
  compute_eigenimages [--help | -h | options] <images>

Arguments:
  images  .npy file containing images as ndarray of dimensions (ni, ny, nx)

Options:
  -n <nim> , --nimgs=<nim>   number of images to use (< 0: all) [default: -1]
  -e <err> , --imerr=<err>   image approximation error tolerance (<= 0: not set,
                             run in interactive mode) [default: 0]
  -a <arch>, --arch=<arch>   architecture [default: cpu]
'''

try:
    from docopt import docopt
    __version__ = '0.1.0'
    have_docopt = True
except:
    have_docopt = False

import numpy
import pylab
import sys
import timeit

from raleigh.interfaces.pca import pca

try:
    input = raw_input
except NameError:
    pass


def _norm(a, axis):
    return numpy.apply_along_axis(numpy.linalg.norm, axis, a)


if have_docopt:
    __version__ = '0.1.0'
    args = docopt(__doc__, version=__version__)
    file = args['<images>']
    ni = int(args['--nimgs'])
    err_tol = float(args['--imerr'])
    arch = args['--arch']
else:
    narg = len(sys.argv)
    if narg < 3:
        print('Usage: compute_eigenimages <images> <pca_err> [gpu]')
    file = sys.argv[1]
    err_tol = float(sys.argv[2])
    arch = 'cpu' if narg < 4 else 'gpu!'
    ni = -1

numpy.random.seed(1) # make results reproducible

all_images = numpy.load(file)

m_all, ny, nx = all_images.shape
n = nx*ny

if ni < 0 or ni > m_all:
    ni = m_all

if ni < m_all:
    print('using first %d images only...' % ni)
    m = ni
    images = all_images[:m,:,:]
else:
    print('using all %d images...' % m_all)
    m = m_all
    images = all_images

vmin = numpy.amin(images)
vmax = numpy.amax(images)
print('data range: %e to %e' % (vmin, vmax))

images = numpy.reshape(images, (m, n))

start = timeit.default_timer()
mean, coord, eigim = pca(images, tol=err_tol, arch=arch, verb=1)
elapsed_time = timeit.default_timer() - start

sigma = _norm(coord, axis=0)
ncon = sigma.shape[0]
print('%d eigenimages computed' % ncon)
if err_tol > 0:
    print('elapsed time: %.2e' % elapsed_time)

while True:
    i = int(input('image number (negative to exit): '))
    if i < 0 or i >= m:
        break
    pylab.figure()
    pylab.title('image %d' % i)
    img = images[i,:]
    image = numpy.reshape(img, (ny, nx))
    pylab.imshow(image, cmap='gray')
    img = numpy.dot(coord[i,:], eigim) + mean
    pca_image = numpy.reshape(img, (ny, nx))
    pylab.figure()
    pylab.title('PCA approximation of the image')
    pylab.imshow(pca_image, cmap='gray')
    pylab.show()

print('saving...')
mean = numpy.reshape(mean, (ny, nx))
eigim = numpy.reshape(eigim, (ncon, ny, nx))
numpy.savez('eigenimages', eigim=eigim, coord=coord, mean=mean)

print('done')



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

# -*- coding: utf-8 -*-
"""Truncated Singular Value Decomposition of a matrix 
   represented by a 2D ndarray.
"""

import math
import numpy
import numpy.linalg as nla
import time

from ..core.solver import Options
from ..algebra.dense_matrix import AMatrix
from .partial_svd import PartialSVD

try:
    input = raw_input
except NameError:
    pass


def truncated_svd(matrix, opt=Options(), nsv=-1, tol=-1, norm='s', msv=-1, \
                  vtol=1e-3, arch='cpu', verb=0):
    '''Computes truncated Singular Value Decomposition of a dense matrix A.
    
    For a given m by n matrix A computes m by k matrix U, k by k diagonal
    matrix S and n by k matrix V such that A V = U S, the columns of U and V
    are orthonirmal (orthogonal and of unit norm) and the largest singular
    value of A - U S V' is smallest possible for a given k (V' = V.T for a real
    A and A.T.conj() for a complex A).
    The diagonal entries of S are the largest k singular values of A and the
    columns of U and V are corresponding left and right singular vectors.

    Parameters
    ----------
    matrix : 2D numpy array
        Matrix A.
    opt : an object of class raleigh.solver.Options
        Solver options (see raleigh.solver).
    nsv : int
        Required number of singular values and vectors if known.
        If negative, implicitely defined by the required truncation tolerance
        or interactively by the user.
    tol : float
        Truncation tolerance in the case nsv < 0. If tol is non-zero, then the 
        computation of singular values and vectors will stop when the norm of 
        D = A - U S V' becomes not greater than eps, where eps is the norm of 
        A multiplied by tol if tol > 0 and eps = -tol if tol < 0. If tol is 
        zero, then the user will be asked repeatedly whether the computation
        should continue (the number of computed singular values and the 
        relative truncation error, the ratio of the norm of D to that of A, are
        displayed).
    norm : character
        The norm to be used for evaluating the truncation error:
        's' : the largest singular value of D,
        'f' : Frobenius norm of D,
        'm' : the largest norm of a row of D.
    msv : int
        Maximal number of singular values to compute. Ignored if negative, 
        otherwise if msv < min(m, n), then the truncation error may be greater
        than requested.
    vtol : float
        Singular vectors error tolerance.
    arch : string
        'cpu' : run on CPU,
        'gpu' : run on GPU if available, otherwise on CPU,
        'gpu!' : run on GPU, throw RuntimError if GPU is not present.
    verb : int
        Verbosity level.

    Returns
    -------
    u : numpy array of shape (m, k)
        The matrix U.
    sigma : numpy array of shape (k,)
        The array of the largest k singular values in descending order.
    vt : numpy array of shape (k, n)
        The matrix V'.

    Notes
    -----
    This solver can only be efficient if singular values decay quickly, e.g.
    exponentially. If the singular values properties are unknown, then it is
    worth to try it first in interactive mode (nsv < 0, tol = 0), and if the
    computation is too slow, use scipy.linalg.svd instead.
    '''
    matrix = AMatrix(matrix, arch=arch)
    psvd = PartialSVD(matrix)

    user_bs = opt.block_size
    if user_bs < 1 and (nsv < 0 or nsv > 100):
        opt.block_size = 128
    if opt.convergence_criteria is None:
        no_cc = True
        opt.convergence_criteria = _DefaultSVDConvergenceCriteria(vtol)
    else:
        no_cc = False
    if opt.stopping_criteria is None and nsv < 0:
        no_sc = True
        opt.stopping_criteria = \
            DefaultStoppingCriteria(matrix, tol, norm, msv, verb)
        v = psvd.vectors()
        opSVD = psvd.op_svd()
        opt.stopping_criteria.err_calc.set_up(opSVD, v, shift=False)
    else:
        no_sc = False

    psvd.compute(matrix, opt, nsv=(0, nsv))
    u = psvd.left()
    v = psvd.right()
    sigma = psvd.sigma
    if msv > 0 and u.shape[1] > msv:
        u = u[:, : msv]
        v = v[:, : msv]
        sigma = sigma[: msv]

    # restore user opt to avoid side effects
    opt.block_size = user_bs
    if no_cc:
        opt.convergence_criteria = None
    if no_sc:
        opt.stopping_criteria = None

    return u, sigma, v.T


class TruncatedSVDErrorCalculator:
    def __init__(self, a):
        m, n = a.shape()
        self.dt = a.data_type()
        s = a.dots()
        self.norms = numpy.sqrt(s.reshape((m, 1)))
        self.solver = None
        self.err = None
        self.op = None
        self.m = m
        self.n = n
        self.shift = False
        self.ncon = 0
        self.err = self.norms.copy()
        self.aves = None
    def set_up(self, op, eigenvectors, shift=False):
        self.op = op.op
        self.eigenvectors = eigenvectors
        self.shift = shift
        if shift:
            self.ones = op.ones
            self.aves = op.aves
            s = self.aves.dots(self.aves)
            vb = eigenvectors.new_vectors(1, self.m)
            self.op.apply(self.aves, vb)
            b = vb.data().reshape((self.m, 1))
            t = (self.norms*self.norms).reshape((self.m, 1))
            x = t - 2*b + s*numpy.ones((self.m, 1))
            self.err = numpy.sqrt(abs(x))
        self.err_init = numpy.amax(self.err)
        self.err_init_f = nla.norm(self.err)
    def update_errors(self):
        ncon = self.eigenvectors.nvec()
        new = ncon - self.ncon
        if new > 0:
            err = self.err*self.err 
            x = self.eigenvectors
            sel = x.selected()
            x.select(new, self.ncon)
            m = self.m
            n = self.n
            if m < n:
                z = x.new_vectors(new, n)
                self.op.apply(x, z, transp=True)
                if self.shift:
                    s = x.dot(self.ones)
                    z.add(self.aves, -1, s)
                y = x.new_vectors(new, m)
                self.op.apply(z, y)
                if self.shift:
                    s = z.dot(self.aves)
                    y.add(self.ones, -1, s)
                q = x.dots(y, transp=True)
                q[q < 0] = 0
                err[q <= 0] = 0
            else:
                y = x.new_vectors(new, m)
                self.op.apply(x, y)
                if self.shift:
                    s = y.dot(self.ones)
                    y.add(self.ones, -1.0/m, s)
                    # accurate orthogonalization needed!
                    s = y.dot(self.ones)
                    y.add(self.ones, -1.0/m, s)
                q = y.dots(y, transp=True)
            q = q.reshape((m, 1))
            err -= q
            err[err < 0] = 0
            self.err = numpy.sqrt(err)
            self.eigenvectors.select(sel[1], sel[0])
            self.ncon = ncon
        return self.err


class DefaultStoppingCriteria:

    def __init__(self, a, err_tol=0, norm='f', max_nsv=0, verb=0):
        self.shape = a.shape()
        self.scale = a.scale()
        self.err_tol = err_tol
        self.norm = norm
        self.max_nsv = max_nsv
        self.verb = verb
        self.ncon = 0
        self.sigma = 1
        self.iteration = 0
        self.start_time = time.time()
        self.elapsed_time = 0
        self.err_calc = TruncatedSVDErrorCalculator(a)
        self.norms = self.err_calc.norms
        self.max_norm = numpy.amax(self.norms)
        self.f_norm = math.sqrt(numpy.sum(self.norms*self.norms))
        self.f = 0

    def satisfied(self, solver):
        self.norms = self.err_calc.norms
        m, n = self.shape
#        scale_max = self.scale*math.sqrt(n)
#        scale_f = self.scale*math.sqrt(m*n)
        scale_max = self.err_calc.err_init
        scale_f = self.err_calc.err_init_f
        if solver.rcon <= self.ncon:
            return False
        new = solver.rcon - self.ncon
        lmd = solver.eigenvalues[self.ncon : solver.rcon]
        sigma = -numpy.sort(-numpy.sqrt(abs(lmd)))
        if self.ncon == 0:
            self.sigma = sigma[0]
            self.err = self.err_calc.err
            self.f = numpy.sum(self.err*self.err)
        i = new - 1
        si = sigma[i]
        si_rel = si/self.sigma
        if self.norm == 'm':
            self.err = self.err_calc.update_errors()
            err_abs = numpy.amax(self.err)
            err_rel = err_abs/scale_max
#            err_rel = err_abs/self.max_norm
        elif self.norm == 'f':
            self.f -= numpy.sum(sigma*sigma)
            err_abs = math.sqrt(abs(self.f))
            err_rel = err_abs/scale_f
#            err_rel = err_abs/self.f_norm
        else:
            err_abs = si
            err_rel = si_rel
        now = time.time()
        elapsed_time = now - self.start_time
        self.elapsed_time += elapsed_time
        if self.norm in ['f', 'm']:
            msg = '%.2f sec: sigma[%d] = %.2e*sigma[0], truncation error = %.2e' % \
                  (self.elapsed_time, self.ncon + i, si_rel, err_rel)
        else:
            msg = '%.2f sec: sigma[%d] = %e = %.2e*sigma[0]' % \
                (self.elapsed_time, self.ncon + i, si, si_rel)
        self.ncon = solver.rcon
        done = False
        if self.err_tol != 0:
            if self.verb > 0:
                print(msg)
            if self.err_tol > 0:
                done = err_rel <= self.err_tol
            else:
                done = err_abs <= abs(self.err_tol)
        elif self.max_nsv < 1:
            done = (input(msg + ', more? ') == 'n')
        elif self.verb > 0:
            print(msg)
        self.iteration = solver.iteration
        self.start_time = time.time()
        done = done or self.max_nsv > 0 and self.ncon >= self.max_nsv
        return done


class _DefaultSVDConvergenceCriteria:
    def __init__(self, tol):
        self.tolerance = tol
    def set_tolerance(self, tolerance):
        self.tolerance = tolerance
    def satisfied(self, solver, i):
        err = solver.convergence_data('kinematic vector error', i)
        return err >= 0 and err <= self.tolerance



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""Reads images from an lfw-style folder (visit http://vis-www.cs.umass.edu/lfw),
   optionally processes them and saves in .npy file.

   The main purpose of processing is to bring images closer to passport format,
   the ideal showcase for PCA as one achieves a dramatic reduction in data size.
   Additionally, the original data set can be doubled in size by adding mirror
   images for the sake of testing on larger data.

Usage:
  convert_lfw [--help | -h | options] <datapath>

Arguments:
  datapath             lfw images folder

Options:
  -m <m>, --how-many=<m>      number of images to process (<0: all)
                              [default: -1]
  -o <file>, --output=<file>  output file name [default: images.npy]
  -s <asym>, --asymm=<asym>   select images that differ from their mirror ones
                              by no greater than s times maximal difference
                              if s > 0, or mean difference otherwise, and save
                              them to photos.npy [default: 1.0]
  -f <v>, --off-face=<v>      if v >= 0, pixel values outside the face area
                              will be set to v times average value, erasing
                              most of the background to get closer to passport
                              photo look (PCA showcase) [default: -1]
  -d, --double         double the number of images by adding mirror images
  -v, --view           view processed images
"""

try:
    from docopt import docopt
    __version__ = '0.1.0'
    have_docopt = True
except:
    have_docopt = False

import numpy
import os
import pylab
import scipy.ndimage as ndimage
import sys

try:
    input = raw_input
except NameError:
    pass


def _mask(nx, ny):
    mask = numpy.zeros((ny, nx), dtype=numpy.uint8)
    x0 = nx/2
    y0 = ny/2
#    ax = nx/5
#    ay = ny/4
    ax = x0 - nx/5
    ay = y0 - ny/6
#    ay = y0 - ny/10
    for y in range(ny):
        for x in range(nx):
            if ((x - x0)/ax)**2 + ((y - y0)/ay)**2 > 1:
                mask[y, x] = 1
    return mask


if have_docopt:
    __version__ = '0.1.0'
    args = docopt(__doc__, version=__version__)
#    print(args)
    datapath = args['<datapath>']
    m = int(args['--how-many'])
    output = args['--output']
    asymm = float(args['--asymm'])
    qb = float(args['--off-face'])
    dble = args['--double']
    view = args['--view']
else:
    narg = len(sys.argv)
    if narg < 2:
        print('Usage: convert_lfw <lfw_folder>')
    datapath = sys.argv[1]
    m = -1
    output = 'images.npy'
    asymm = 1.0
    dble = False
    view = False
    qb = -1

face = (qb >= 0)
if face:
    qx = 0.7
    qy = 0.9
else:
    qx = 1.0
    qy = 1.0

print('processing images from %s...' % datapath)

image_dir = datapath
for subdir in os.listdir(image_dir):
    if subdir.endswith('.txt'):
        continue
    fulldir = image_dir + '/' + subdir
    for filename in os.listdir(fulldir):
        if not filename.endswith('.jpg'):
            continue
        fullname = fulldir + '/' + filename
        image = ndimage.imread(fullname, mode = 'L').astype(numpy.float32)
        break
    break

ny, nx = image.shape
ix = int(nx*(1 - qx)/2)
iy = int(ny*(1 - qy)/2)
nx = int(nx*qx)
ny = int(ny*qy)
jx = ix + nx
jy = iy + ny

print('counting images...')
ndir = 0
nimg = 0
for subdir in os.listdir(image_dir):
    if subdir.endswith('.txt'):
        continue
    fulldir = image_dir + '/' + subdir
    for filename in os.listdir(fulldir):
        if not filename.endswith('.jpg'):
            continue
        fullname = fulldir + '/' + filename
        nimg += 1
        if m > 0 and nimg >= m:
            break
    ndir += 1
    if m > 0 and nimg >= m:
        break

if dble:
    ni = 2*nimg
else:
    ni = nimg
images = numpy.zeros((ni, ny, nx), dtype=numpy.float32)
names = numpy.ndarray((ni,), dtype=object)
offsets = numpy.zeros((ni,), dtype=int)

print('collecting %d images from %d folders...' % (nimg, ndir))
ndir = 0
nimg = 0
for subdir in os.listdir(image_dir):
    if subdir.endswith('.txt'):
        continue
    fulldir = image_dir + '/' + subdir
    for filename in os.listdir(fulldir):
        if not filename.endswith('.jpg'):
            continue
        fullname = fulldir + '/' + filename
        image = ndimage.imread(fullname, mode='L')
        if dble:
            images[2*nimg, :, :] = image[iy : jy, ix : jx]
            names[2*nimg] = subdir
            names[2*nimg + 1] = subdir
        else:
            images[nimg, :, :] = image[iy : jy, ix : jx]
            names[nimg] = subdir
        nimg += 1
        if m > 0 and nimg >= m:
            break
    ndir += 1
    if m > 0 and nimg >= m:
        break

file = open("names.txt", "w+")
for i in range(ni):
    file.write('%s\n' % names[i])
file.close()

vmax = numpy.amax(images)
vmin = numpy.amin(images)
print('pixel values range: %f to %f' % (vmin, vmax))

if asymm < 1.0:
    a = numpy.zeros(ni)
    ia = numpy.zeros(ni, dtype=numpy.int32)
mask = _mask(nx, ny)
n = nx*ny
for i in range(nimg):
    if dble:
        j = 2*i
    else:
        j = i
    image = images[j,:,:]
    if face:
        image[mask > 0] = vmin + qb*(vmax - vmin)
#        image[mask > 0] = 0
#        v = numpy.sum(image)/(n - numpy.sum(mask))
#        image[mask > 0] = v
        images[j,:,:] = image
    if asymm < 1.0:
        img = numpy.reshape(image, (n,))
        gmi = numpy.reshape(image[:, ::-1], (n,))
        a[j] = numpy.linalg.norm(img - gmi)/numpy.linalg.norm(img)
        ia[j] = j
    if dble:
        images[j + 1, :, :] = image[:, ::-1]
        if asymm < 1.0:
            a[j + 1] = a[j]
            ia[j + 1] = j + 1

if asymm < 1.0:
    ind = numpy.argsort(a)
    pylab.figure()
    pylab.plot(numpy.arange(1, ni + 1, 1), a[ind])
    pylab.show()
    max_asymm = numpy.amax(a)
    mean_asymm = numpy.mean(a)
    print('asymmetry: max %f, mean %f' % (max_asymm, mean_asymm))
    if asymm > 0:
        th = max_asymm*asymm
    else:
        th = mean_asymm
    k = sum(a <= th)
    iind = numpy.sort(ia[ind[:k]])
    photos = images[iind,:,:]
    print(photos.shape)
    while True:
        i = int(input('photo-image: '))
        if i < 0 or i >= k:
            break
        image = photos[i,:,:]
        pylab.title('%s' % names[iind[i]].replace('_', ' '))
        pylab.imshow(image, cmap='gray', vmin=vmin, vmax=vmax)
        pylab.show()
    print('saving %d photos to %s...' % (k, 'photos.npy'))
    numpy.save('photos.npy', photos)
    file = open("photo_names.txt", "w+")
    for i in range(k):
        j = iind[i]
        file.write('%s\n' % names[j])
    file.close()

print('saving %d images to %s...' % (images.shape[0], output))
numpy.save(output, images)

while view:
    i = int(input('image: '))
    if i < 0 or i >= images.shape[0]:
        break
    image = images[i,:,:]
    pylab.figure()
    pylab.title('%s' % names[i].replace('_', ' '))
    pylab.imshow(image, cmap='gray', vmin=vmin, vmax=vmax)
    pylab.show()

print('done')



'''RALEIGH: RAL EIGensolver for real symmetric and Hermitian problems.

core:
    Contains core eigensolver routine implemented in terms of an abstract
    Vectors class.
    Direct use is for advanced users only - start by trying user-friendly
    interfaces in ./interfaces.

algebra:
    Implementations of Vectors class (and related matrix classes) on CPU and
    GPU.

interfaces:
    User-friendly interfaces to solver.

examples:
    Example scripts demonstrating the use of interfaces for the solution of
    eigenvalue problems for sparse matrices and singular value problems for
    dense matrices.
'''



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""CPU dense algebra selector.
"""

from . import verbosity


try:
    from .dense_cblas import Vectors, Matrix
    if verbosity.level > 0:
        print('using mkl cblas...')
except:
    if verbosity.level > 0:
        print('mkl cblas not found, using numpy...')
    from .dense_numpy import Vectors, Matrix



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""Partial eigenvalue problem solver for a sparse symmetric/Hermitian matrix.

--------------------------------------------------------------------------------
Requires MKL 10.3 or later (needs mkl_rt.dll on Windows, libmkl_rt.so on Linux).
--------------------------------------------------------------------------------
"""

import numpy
import time

from ..algebra.dense_cblas import Vectors
from ..algebra.sparse_mkl import SparseSymmetricMatrix
from ..algebra.sparse_mkl import SparseSymmetricSolver
from ..algebra.sparse_mkl import Operator
from ..core.solver import Problem, Solver, Options, DefaultConvergenceCriteria


def partial_hevp(A, B=None, T=None, sigma=0, which=6, tol=1e-4, verb=0):
    '''Computes several eigenpairs of sparse real symmetric/Hermitian eigenvalue
    problems using either shift-invert or preconditioning technique.
    Requires MKL 10.3 or later (needs mkl_rt.dll on Windows, libmkl_rt.so on
    Linux).

    Parameters
    ----------
    A : either scipy's sparse matrix or raleigh's SparseSymmetricSolver
        In the first case, the (stiffness) matrix of the problem.
        In the second case, the inverse of A - sigma B if B is not None or
        A - sigma I, where I is the identity, otherwise.
    B : scipy's sparse matrix
        Mass matrix. If None, eigenvalues and eigenvectors of A are computed,
        otherwise those of the generalized problem A x = lambda B x.
    T : a Python object
        If T is not None, then A must be positive definite.
        Preconditioner (roughly, approximate inverse of A). Must have method
        apply(x, y) that, for a given equally shaped 2D ndarrays x and y with
        the second dimension equal to the problem size, applies preconditioning
        to rows of x and places the results into respective rows of y.
        The method apply(x, y) must act as a self-adjoint positive definite 
        linear operator, i.e. for any x, the matrix numpy.dot(x, y), where y is
        computed by apply(x, y), must be real symmetric/Hermitian and positive
        definite.
    sigma : float
        Ignored if T is not None. Otherwise specifies value in the vicinity of
        which the wanted eigenvalues are situated.
    which : an integer or tuple of integers
        Specifies which eigenvalues are wanted. If T is not none, then it is
        the number of smallest eigenvalues. Otherwise, if it is an integer k,
        then k eigenvalues nearest to sigma will be computed, and if it is a
        tuple (k, l), then k neares eigenvalues left from sigma and l nearest
        eigenvalues right from sigma will be computed.
    tol : float
        Eigenvector error tolerance.
    verb : integer
        Verbosity level.
        < 0 : nothing printed
          0 : error and warning messages printed
          1 : + number of iteration and converged eigenvalues printed
          2 : + current eigenvalue iterates, residuals and error estimates
              printed
    '''

    if B is not None:
        opB = SparseSymmetricMatrix(B)
    else:
        opB = None

    if T is None:
        try:
            n = A.size()
            dtype = A.data_type()
            sigma = A.sigma()
            neg, pos = A.inertia()
            solver = A
        except:
            m, n = A.shape
            if m != n:
                raise ValueError('the matrix must be square')
            dtype = A.data.dtype
            solver = SparseSymmetricSolver(dtype=dtype)
            if verb > -1:
                print('setting up the linear system solver...')
            start = time.time()
            solver.analyse(A, sigma, B)
            solver.factorize()
            stop = time.time()
            setup_time = stop - start
            if verb > -1:
                print('setup time: %.2e' % setup_time)
        opAinv = solver
        neg, pos = solver.inertia()
        if verb > -1:
            print('positive eigenvalues: %d' % pos)
            print('negative eigenvalues: %d' % neg)
        try:
            l = len(which)
            if l != 2:
                raise ValueError\
                      ('which must be either integer or tuple of 2 integers')
            left = min(which[0], neg)
            right = min(which[1], pos)
            which = (left, right)
        except:
            pass
        eigenvectors = Vectors(n, data_type=dtype)
        if B is None:
            evp = Problem(eigenvectors, opAinv)
        else:
            evp = Problem(eigenvectors, opAinv, opB, 'pro')
        evp_solver = Solver(evp)
    else:
        A = SparseSymmetricMatrix(A)
        n = A.size()
        dtype = A.data_type()
        eigenvectors = Vectors(n, data_type=dtype)
        opA = A
        opT = Operator(T)
        if B is None:
            evp = Problem(eigenvectors, opA)
        else:
            evp = Problem(eigenvectors, opA, opB, 'gen')
        evp_solver = Solver(evp)
        evp_solver.set_preconditioner(opT)
        sigma = None
        which = (which, 0)

    opt = Options()
    opt.convergence_criteria = DefaultConvergenceCriteria()
    opt.convergence_criteria.set_error_tolerance('k eigenvector error', tol)
    opt.sigma = sigma
    opt.verbosity = verb

    start = time.time()
    status = evp_solver.solve(eigenvectors, opt, which=which)
    stop = time.time()
    solve_time = stop - start
    if T is None:
        lmd = sigma + 1./evp_solver.eigenvalues
    else:
        lmd = evp_solver.eigenvalues
    ind = numpy.argsort(lmd)
    lmd = lmd[ind]
    ne = eigenvectors.nvec()
    if verb > -1:
        print('iterations: %d, solve time: %.2e' % \
              (evp_solver.iteration, solve_time))
    x = eigenvectors.data().T
    if ne > 0:
        x = x[:, ind]
    return lmd, x, status



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""Partial SVD of a matrix represented by a 2D ndarray.

   For advanced users only.
"""

import math
import numpy
import numpy.linalg as nla
import time

from ..core.solver import Problem, Solver, Options


class PartialSVD:

    def __init__(self, matrix, shift=False):
        op = matrix.as_operator()
        m, n = matrix.shape()
        transp = m < n
        if transp:
            n, m = m, n
        v = op.new_vectors(n)
        dt = v.data_type()
        opSVD = _OperatorSVD(matrix, v, transp, shift)
        self.__op = op
        self.__transp = transp
        self.__shape = (m, n)
        self.__shift = shift
        self.__v = v
        self.__dtype = dt
        self.__opsvd = opSVD
        self.sigma = None
        self.__left = None
        self.__right = None
        self.__mean = None
        self.__left_v = None
        self.__right_v = None
        self.__mean_v = None
        self.iterations = -1

    def op_svd(self):
        return self.__opsvd

    def vectors(self):
        return self.__v

    def compute(self, matrix, opt=Options(), nsv=(-1, -1), refine=True):
    
        op = self.__op
        m, n = self.__shape
        transp = self.__transp
        v = self.__v
        dt = self.__dtype
        opSVD = self.__opsvd
        shift = self.__shift

        problem = Problem(v, opSVD)
        solver = Solver(problem)

        solver.solve(v, options=opt, which=nsv)
        if opt.verbosity > 0:
            print('operator application time: %.2e' % opSVD.time)

        nv = v.nvec()
        u = v.new_vectors(nv, m)
        if nv > 0:
            op.apply(v, u, transp)
            if shift:
                m, n = op.shape()
                dt = op.data_type()
                ones = numpy.ones((1, m), dtype=dt)
                e = v.new_vectors(1, m)
                e.fill(ones)
                w = v.new_vectors(1, n)
                op.apply(e, w, transp=True)
                w.scale(m*ones[0,:1])
                if not transp:
                    s = v.dot(w)
                    u.add(e, -1, s)
                else:
                    s = v.dot(e)
                    u.add(w, -1, s)
            if refine:
                sigma, q = u.svd()
                w = v.new_vectors(nv)
                v.multiply(_conj(q.T), w)
                w.copy(v)
            else:
                sigma = numpy.sqrt(abs(u.dots(u)))
                u.scale(sigma)
                w = u.new_vectors(nv)
                ind = numpy.argsort(-sigma)
                sigma = sigma[ind]
                u.copy(w, ind)
                w.copy(u)
                w = v.new_vectors(nv)
                v.copy(w, ind)
                w.copy(v)
        else:
            sigma = numpy.ndarray((0,), dtype=v.data_type())
        self.sigma = sigma
        self.__mean_v = opSVD.mean_v()
        self.iterations = solver.iteration
        if transp:
            self.__left_v = v
            self.__right_v = u
        else:
            self.__left_v = u
            self.__right_v = v

    def mean(self):
        if self.__mean is None:
            if self.__mean_v is not None:
                self.__mean = self.__mean_v.data()
        return self.__mean

    def left(self):
        if self.__left is None:
            if self.__left_v is not None:
                self.__left = self.__left_v.data().T
        return self.__left

    def right(self):
        if self.__right is None:
            if self.__right_v is not None:
                self.__right = self.__right_v.data().T
        return self.__right

    def mean_v(self):
        return self.__mean_v

    def left_v(self):
        return self.__left_v

    def right_v(self):
        return self.__right_v


class _OperatorSVD:
    def __init__(self, matrix, v, transp=False, shift=False):
        self.op = matrix.as_operator()
        self.gpu = matrix.gpu()
        self.transp = transp
        self.shift = shift
        self.time = 0
        m, n = self.op.shape()
        if transp:
            self.w = v.new_vectors(0, n)
        else:
            self.w = v.new_vectors(0, m)
        if shift:
            dt = self.op.data_type()
            ones = numpy.ones((1, m), dtype=dt)
            self.ones = v.new_vectors(1, m)
            self.ones.fill(ones)
            self.aves = v.new_vectors(1, n)
            self.op.apply(self.ones, self.aves, transp=True)
            self.aves.scale(m*ones[0,:1])
    def apply(self, x, y):
        m, n = self.op.shape()
        k = x.nvec()
        start = time.time()
        if self.transp:
            if self.w.nvec() < k:
                self.w = x.new_vectors(k, n)
            z = self.w
            z.select(k)
            self.op.apply(x, z, transp=True)
            if self.shift:
                s = x.dot(self.ones)
                z.add(self.aves, -1, s)
            self.op.apply(z, y)
            if self.shift:
                s = z.dot(self.aves)
                y.add(self.ones, -1, s)
        else:
            if self.w.nvec() < k:
                self.w = x.new_vectors(k, m)
            z = self.w
            z.select(k)
            self.op.apply(x, z)
            if self.shift:
                s = z.dot(self.ones)
                z.add(self.ones, -1.0/m, s)
                # accurate orthogonalization needed!
                s = z.dot(self.ones)
                z.add(self.ones, -1.0/m, s)
            self.op.apply(z, y, transp=True)
        if self.gpu is not None:
            self.gpu.synchronize()
        stop = time.time()
        self.time += stop - start
    def mean(self):
        if self.shift:
            return self.aves.data()
        else:
            return None
    def mean_v(self):
        if self.shift:
            return self.aves
        else:
            return None


def _norm(a, axis):
    return numpy.apply_along_axis(nla.norm, axis, a)


def _conj(a):
    if a.dtype.kind == 'c':
        return a.conj()
    else:
        return a
